#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from distutils.version import LooseVersion

import matplotlib
import numpy as np
import pandas as pd
from matplotlib.axes._base import _process_plot_format
from pandas.core.dtypes.inference import is_integer, is_list_like
from pandas.io.formats.printing import pprint_thing
from pandas.core.base import PandasObject
from pyspark.ml.feature import Bucketizer
from pyspark.sql import functions as F

from databricks.koalas.missing import _unsupported_function
from databricks.koalas.config import get_option


def _gca(rc=None):
    import matplotlib.pyplot as plt
    with plt.rc_context(rc):
        return plt.gca()


def _get_standard_kind(kind):
    return {'density': 'kde'}.get(kind, kind)


if LooseVersion(pd.__version__) < LooseVersion('0.25'):
    from pandas.plotting._core import _all_kinds, BarPlot, BoxPlot, HistPlot, MPLPlot, PiePlot, \
        AreaPlot, LinePlot, BarhPlot, ScatterPlot
else:
    from pandas.plotting._core import PlotAccessor
    from pandas.plotting._matplotlib import BarPlot, BoxPlot, HistPlot, PiePlot, AreaPlot, \
        LinePlot, BarhPlot, ScatterPlot
    from pandas.plotting._matplotlib.core import MPLPlot
    _all_kinds = PlotAccessor._all_kinds


class TopNPlot:

    def get_top_n(self, data):
        from databricks.koalas import DataFrame, Series
        max_rows = get_option("plotting.max_rows")
        # Simply use the first 1k elements and make it into a pandas dataframe
        # For categorical variables, it is likely called from df.x.value_counts().plot.xxx().
        if isinstance(data, Series):
            data = data.head(max_rows + 1).to_pandas().to_frame()
        elif isinstance(data, DataFrame):
            data = data.head(max_rows + 1).to_pandas()
        else:
            ValueError("Only DataFrame and Series are supported for plotting.")

        self.partial = False
        if len(data) > max_rows:
            self.partial = True
            data = data.iloc[:max_rows]
        return data

    def set_result_text(self, ax):
        max_rows = get_option("plotting.max_rows")
        assert hasattr(self, "partial")

        if self.partial:
            ax.text(1, 1, 'showing top {} elements only'.format(max_rows),
                    size=6, ha='right', va='bottom',
                    transform=ax.transAxes)


class SampledPlot:
    def get_sampled(self, data):
        from databricks.koalas import DataFrame, Series
        fraction = get_option("plotting.sample_ratio")
        if fraction is None:
            fraction = 1 / (len(data) / get_option("plotting.max_rows"))
            fraction = min(1., fraction)
        self.fraction = fraction

        if isinstance(data, DataFrame):
            sampled = data._sdf.sample(fraction=self.fraction)
            return DataFrame(data._internal.copy(sdf=sampled)).to_pandas()
        elif isinstance(data, Series):
            scol = data._scol
            sampled = data._kdf._sdf.sample(fraction=self.fraction)
            return DataFrame(data._kdf._internal.copy(sdf=sampled, scol=scol)).to_pandas()
        else:
            ValueError("Only DataFrame and Series are supported for plotting.")

    def set_result_text(self, ax):
        assert hasattr(self, "fraction")

        if self.fraction < 1:
            ax.text(
                1, 1, 'showing the sampled result by fraction %s' % self.fraction,
                size=6, ha='right', va='bottom',
                transform=ax.transAxes)


class KoalasBarPlot(BarPlot, TopNPlot):
    def __init__(self, data, **kwargs):
        super(KoalasBarPlot, self).__init__(self.get_top_n(data), **kwargs)

    def _plot(self, ax, x, y, w, start=0, log=False, **kwds):
        self.set_result_text(ax)
        return ax.bar(x, y, w, bottom=start, log=log, **kwds)


class KoalasBoxPlotSummary:
    def __init__(self, data, colname):
        self.data = data
        self.colname = colname

    def compute_stats(self, whis, precision):
        # Computes mean, median, Q1 and Q3 with approx_percentile and precision
        pdf = (self.data._kdf._sdf
               .agg(*[F.expr('approx_percentile({}, {}, {})'.format(self.colname, q,
                                                                    1. / precision))
                      .alias('{}_{}%'.format(self.colname, int(q * 100)))
                      for q in [.25, .50, .75]],
                    F.mean(self.colname).alias('{}_mean'.format(self.colname))).toPandas())

        # Computes IQR and Tukey's fences
        iqr = '{}_iqr'.format(self.colname)
        p75 = '{}_75%'.format(self.colname)
        p25 = '{}_25%'.format(self.colname)
        pdf.loc[:, iqr] = pdf.loc[:, p75] - pdf.loc[:, p25]
        pdf.loc[:, '{}_lfence'.format(self.colname)] = pdf.loc[:, p25] - whis * pdf.loc[:, iqr]
        pdf.loc[:, '{}_ufence'.format(self.colname)] = pdf.loc[:, p75] + whis * pdf.loc[:, iqr]

        qnames = ['25%', '50%', '75%', 'mean', 'lfence', 'ufence']
        col_summ = pdf[['{}_{}'.format(self.colname, q) for q in qnames]]
        col_summ.columns = qnames
        lfence, ufence = col_summ['lfence'], col_summ['ufence']

        stats = {'mean': col_summ['mean'].values[0],
                 'med': col_summ['50%'].values[0],
                 'q1': col_summ['25%'].values[0],
                 'q3': col_summ['75%'].values[0]}

        return stats, (lfence.values[0], ufence.values[0])

    def outliers(self, lfence, ufence):
        # Builds expression to identify outliers
        expression = F.col(self.colname).between(lfence, ufence)
        # Creates a column to flag rows as outliers or not
        return self.data._kdf._sdf.withColumn('__{}_outlier'.format(self.colname), ~expression)

    def calc_whiskers(self, outliers):
        # Computes min and max values of non-outliers - the whiskers
        minmax = (outliers
                  .filter('not __{}_outlier'.format(self.colname))
                  .agg(F.min(self.colname).alias('min'),
                       F.max(self.colname).alias('max'))
                  .toPandas())
        return minmax.iloc[0][['min', 'max']].values

    def get_fliers(self, outliers):
        # Filters only the outliers, should "showfliers" be True
        fliers_df = outliers.filter('__{}_outlier'.format(self.colname))

        # If shows fliers, takes the top 1k with highest absolute values
        fliers = (fliers_df
                  .select(F.abs(F.col('`{}`'.format(self.colname))).alias(self.colname))
                  .orderBy(F.desc('`{}`'.format(self.colname)))
                  .limit(1001)
                  .toPandas()[self.colname].values)

        return fliers


class KoalasBoxPlot(BoxPlot):
    @staticmethod
    def rc_defaults(notch=None, vert=None, whis=None,
                    patch_artist=None, bootstrap=None, meanline=None,
                    showmeans=None, showcaps=None, showbox=None,
                    showfliers=None, **kwargs):
        # Missing arguments default to rcParams.
        if whis is None:
            whis = matplotlib.rcParams['boxplot.whiskers']
        if bootstrap is None:
            bootstrap = matplotlib.rcParams['boxplot.bootstrap']

        if notch is None:
            notch = matplotlib.rcParams['boxplot.notch']
        if vert is None:
            vert = matplotlib.rcParams['boxplot.vertical']
        if patch_artist is None:
            patch_artist = matplotlib.rcParams['boxplot.patchartist']
        if meanline is None:
            meanline = matplotlib.rcParams['boxplot.meanline']
        if showmeans is None:
            showmeans = matplotlib.rcParams['boxplot.showmeans']
        if showcaps is None:
            showcaps = matplotlib.rcParams['boxplot.showcaps']
        if showbox is None:
            showbox = matplotlib.rcParams['boxplot.showbox']
        if showfliers is None:
            showfliers = matplotlib.rcParams['boxplot.showfliers']

        return dict(whis=whis, bootstrap=bootstrap, notch=notch, vert=vert,
                    patch_artist=patch_artist, meanline=meanline, showmeans=showmeans,
                    showcaps=showcaps, showbox=showbox, showfliers=showfliers)

    def boxplot(self, ax, bxpstats, notch=None, sym=None, vert=None,
                whis=None, positions=None, widths=None, patch_artist=None,
                bootstrap=None, usermedians=None, conf_intervals=None,
                meanline=None, showmeans=None, showcaps=None,
                showbox=None, showfliers=None, boxprops=None,
                labels=None, flierprops=None, medianprops=None,
                meanprops=None, capprops=None, whiskerprops=None,
                manage_xticks=True, autorange=False, zorder=None,
                precision=None):

        def _update_dict(dictionary, rc_name, properties):
            """ Loads properties in the dictionary from rc file if not already
            in the dictionary"""
            rc_str = 'boxplot.{0}.{1}'
            if dictionary is None:
                dictionary = dict()
            for prop_dict in properties:
                dictionary.setdefault(prop_dict,
                                      matplotlib.rcParams[rc_str.format(rc_name, prop_dict)])
            return dictionary

        # Common property dictionnaries loading from rc
        flier_props = ['color', 'marker', 'markerfacecolor', 'markeredgecolor',
                       'markersize', 'linestyle', 'linewidth']
        default_props = ['color', 'linewidth', 'linestyle']

        boxprops = _update_dict(boxprops, 'boxprops', default_props)
        whiskerprops = _update_dict(whiskerprops, 'whiskerprops',
                                    default_props)
        capprops = _update_dict(capprops, 'capprops', default_props)
        medianprops = _update_dict(medianprops, 'medianprops', default_props)
        meanprops = _update_dict(meanprops, 'meanprops', default_props)
        flierprops = _update_dict(flierprops, 'flierprops', flier_props)

        if patch_artist:
            boxprops['linestyle'] = 'solid'
            boxprops['edgecolor'] = boxprops.pop('color')

        # if non-default sym value, put it into the flier dictionary
        # the logic for providing the default symbol ('b+') now lives
        # in bxp in the initial value of final_flierprops
        # handle all of the `sym` related logic here so we only have to pass
        # on the flierprops dict.
        if sym is not None:
            # no-flier case, which should really be done with
            # 'showfliers=False' but none-the-less deal with it to keep back
            # compatibility
            if sym == '':
                # blow away existing dict and make one for invisible markers
                flierprops = dict(linestyle='none', marker='', color='none')
                # turn the fliers off just to be safe
                showfliers = False
            # now process the symbol string
            else:
                # process the symbol string
                # discarded linestyle
                _, marker, color = _process_plot_format(sym)
                # if we have a marker, use it
                if marker is not None:
                    flierprops['marker'] = marker
                # if we have a color, use it
                if color is not None:
                    # assume that if color is passed in the user want
                    # filled symbol, if the users want more control use
                    # flierprops
                    flierprops['color'] = color
                    flierprops['markerfacecolor'] = color
                    flierprops['markeredgecolor'] = color

        # replace medians if necessary:
        if usermedians is not None:
            if (len(np.ravel(usermedians)) != len(bxpstats) or
                    np.shape(usermedians)[0] != len(bxpstats)):
                raise ValueError('usermedians length not compatible with x')
            else:
                # reassign medians as necessary
                for stats, med in zip(bxpstats, usermedians):
                    if med is not None:
                        stats['med'] = med

        if conf_intervals is not None:
            if np.shape(conf_intervals)[0] != len(bxpstats):
                err_mess = 'conf_intervals length not compatible with x'
                raise ValueError(err_mess)
            else:
                for stats, ci in zip(bxpstats, conf_intervals):
                    if ci is not None:
                        if len(ci) != 2:
                            raise ValueError('each confidence interval must '
                                             'have two values')
                        else:
                            if ci[0] is not None:
                                stats['cilo'] = ci[0]
                            if ci[1] is not None:
                                stats['cihi'] = ci[1]

        artists = ax.bxp(bxpstats, positions=positions, widths=widths,
                         vert=vert, patch_artist=patch_artist,
                         shownotches=notch, showmeans=showmeans,
                         showcaps=showcaps, showbox=showbox,
                         boxprops=boxprops, flierprops=flierprops,
                         medianprops=medianprops, meanprops=meanprops,
                         meanline=meanline, showfliers=showfliers,
                         capprops=capprops, whiskerprops=whiskerprops,
                         manage_xticks=manage_xticks, zorder=zorder)
        return artists

    def _plot(self, ax, bxpstats, column_num=None, return_type='axes', **kwds):
        bp = self.boxplot(ax, bxpstats, **kwds)

        if return_type == 'dict':
            return bp, bp
        elif return_type == 'both':
            return self.BP(ax=ax, lines=bp), bp
        else:
            return ax, bp

    def _compute_plot_data(self):
        colname = self.data.name
        summary = KoalasBoxPlotSummary(self.data, colname)

        # Updates all props with the rc defaults from matplotlib
        self.kwds.update(KoalasBoxPlot.rc_defaults(**self.kwds))

        # Gets some important kwds
        showfliers = self.kwds.get('showfliers', False)
        whis = self.kwds.get('whis', 1.5)
        labels = self.kwds.get('labels', [colname])

        # This one is Koalas specific to control precision for approx_percentile
        precision = self.kwds.get('precision', 0.01)

        # # Computes mean, median, Q1 and Q3 with approx_percentile and precision
        col_stats, col_fences = summary.compute_stats(whis, precision)

        # # Creates a column to flag rows as outliers or not
        outliers = summary.outliers(*col_fences)

        # # Computes min and max values of non-outliers - the whiskers
        whiskers = summary.calc_whiskers(outliers)

        if showfliers:
            fliers = summary.get_fliers(outliers)
        else:
            fliers = []

        # Builds bxpstats dict
        stats = []
        item = {'mean': col_stats['mean'],
                'med': col_stats['med'],
                'q1': col_stats['q1'],
                'q3': col_stats['q3'],
                'whislo': whiskers[0],
                'whishi': whiskers[1],
                'fliers': fliers,
                'label': labels[0]}
        stats.append(item)

        self.data = {labels[0]: stats}

    def _make_plot(self):
        bxpstats = list(self.data.values())[0]
        ax = self._get_ax(0)
        kwds = self.kwds.copy()

        for stats in bxpstats:
            if len(stats['fliers']) > 1000:
                stats['fliers'] = stats['fliers'][:1000]
                ax.text(1, 1, 'showing top 1,000 fliers only', size=6, ha='right', va='bottom',
                        transform=ax.transAxes)

        ret, bp = self._plot(ax, bxpstats, column_num=0,
                             return_type=self.return_type, **kwds)
        self.maybe_color_bp(bp)
        self._return_obj = ret

        labels = [l for l, _ in self.data.items()]
        labels = [pprint_thing(l) for l in labels]
        if not self.use_index:
            labels = [pprint_thing(key) for key in range(len(labels))]
        self._set_ticklabels(ax, labels)


class KoalasHistPlotSummary:
    def __init__(self, data, colname):
        self.data = data
        self.colname = colname

    def get_bins(self, n_bins):
        boundaries = (self.data._kdf._sdf
                      .agg(F.min(self.colname),
                           F.max(self.colname))
                      .rdd
                      .map(tuple)
                      .collect()[0])
        # divides the boundaries into bins
        return np.linspace(boundaries[0], boundaries[1], n_bins + 1)

    def calc_histogram(self, bins):
        bucket_name = '__{}_bucket'.format(self.colname)
        # creates a Bucketizer to get corresponding bin of each value
        bucketizer = Bucketizer(splits=bins,
                                inputCol=self.colname,
                                outputCol=bucket_name,
                                handleInvalid="skip")
        # after bucketing values, groups and counts them
        result = (bucketizer
                  .transform(self.data._kdf._sdf)
                  .select(bucket_name)
                  .groupby(bucket_name)
                  .agg(F.count('*').alias('count'))
                  .toPandas()
                  .sort_values(by=bucket_name))

        # generates a pandas DF with one row for each bin
        # we need this as some of the bins may be empty
        indexes = pd.DataFrame({bucket_name: np.arange(0, len(bins) - 1),
                                'bucket': bins[:-1]})
        # merges the bins with counts on it and fills remaining ones with zeros
        data = indexes.merge(result, how='left', on=[bucket_name]).fillna(0)[['count']]
        data.columns = [bucket_name]

        return data


class KoalasHistPlot(HistPlot):
    def _args_adjust(self):
        if is_integer(self.bins):
            summary = KoalasHistPlotSummary(self.data, self.data.name)
            # computes boundaries for the column
            self.bins = summary.get_bins(self.bins)

        if is_list_like(self.bottom):
            self.bottom = np.array(self.bottom)

    @classmethod
    def _plot(cls, ax, y, style=None, bins=None, bottom=0, column_num=0,
              stacking_id=None, **kwds):
        if column_num == 0:
            cls._initialize_stacker(ax, stacking_id, len(bins) - 1)

        base = np.zeros(len(bins) - 1)
        bottom = bottom + \
            cls._get_stacked_values(ax, stacking_id, base, kwds['label'])

        # Since the counts were computed already, we use them as weights and just generate
        # one entry for each bin
        n, bins, patches = ax.hist(bins[:-1], bins=bins, bottom=bottom, weights=y, **kwds)

        cls._update_stacker(ax, stacking_id, n)
        return patches

    def _compute_plot_data(self):
        summary = KoalasHistPlotSummary(self.data, self.data.name)
        # generates a pandas DF with one row for each bin
        self.data = summary.calc_histogram(self.bins)


class KoalasPiePlot(PiePlot, TopNPlot):

    def __init__(self, data, **kwargs):
        super(KoalasPiePlot, self).__init__(self.get_top_n(data), **kwargs)

    def _make_plot(self):
        self.set_result_text(self._get_ax(0))
        super(KoalasPiePlot, self)._make_plot()


class KoalasAreaPlot(AreaPlot, SampledPlot):
    def __init__(self, data, **kwargs):
        super(KoalasAreaPlot, self).__init__(self.get_sampled(data), **kwargs)

    def _make_plot(self):
        self.set_result_text(self._get_ax(0))
        super(KoalasAreaPlot, self)._make_plot()


class KoalasLinePlot(LinePlot, SampledPlot):
    def __init__(self, data, **kwargs):
        super(KoalasLinePlot, self).__init__(self.get_sampled(data), **kwargs)

    def _make_plot(self):
        self.set_result_text(self._get_ax(0))
        super(KoalasLinePlot, self)._make_plot()


class KoalasBarhPlot(BarhPlot, TopNPlot):

    def __init__(self, data, **kwargs):
        super(KoalasBarhPlot, self).__init__(self.get_top_n(data), **kwargs)

    def _make_plot(self):
        self.set_result_text(self._get_ax(0))
        super(KoalasBarhPlot, self)._make_plot()


class KoalasScatterPlot(ScatterPlot, TopNPlot):

    def __init__(self, data, x, y, **kwargs):
        super().__init__(self.get_top_n(data), x, y, **kwargs)

    def _make_plot(self):
        self.set_result_text(self._get_ax(0))
        super(KoalasScatterPlot, self)._make_plot()


_klasses = [
    KoalasHistPlot,
    KoalasBarPlot,
    KoalasBoxPlot,
    KoalasPiePlot,
    KoalasAreaPlot,
    KoalasLinePlot,
    KoalasBarhPlot,
    KoalasScatterPlot,
]
_plot_klass = {getattr(klass, '_kind'): klass for klass in _klasses}


def plot_series(data, kind='line', ax=None,                    # Series unique
                figsize=None, use_index=True, title=None, grid=None,
                legend=False, style=None, logx=False, logy=False, loglog=False,
                xticks=None, yticks=None, xlim=None, ylim=None,
                rot=None, fontsize=None, colormap=None, table=False,
                yerr=None, xerr=None,
                label=None, secondary_y=False,                 # Series unique
                **kwds):
    """
    Make plots of Series using matplotlib / pylab.

    Each plot kind has a corresponding method on the
    ``Series.plot`` accessor:
    ``s.plot(kind='line')`` is equivalent to
    ``s.plot.line()``.

    Parameters
    ----------
    data : Series

    kind : str
        - 'line' : line plot (default)
        - 'bar' : vertical bar plot
        - 'barh' : horizontal bar plot
        - 'hist' : histogram
        - 'box' : boxplot
        - 'kde' : Kernel Density Estimation plot
        - 'density' : same as 'kde'
        - 'area' : area plot
        - 'pie' : pie plot

    ax : matplotlib axes object
        If not passed, uses gca()
    figsize : a tuple (width, height) in inches
    use_index : boolean, default True
        Use index as ticks for x axis
    title : string or list
        Title to use for the plot. If a string is passed, print the string at
        the top of the figure. If a list is passed and `subplots` is True,
        print each item in the list above the corresponding subplot.
    grid : boolean, default None (matlab style default)
        Axis grid lines
    legend : False/True/'reverse'
        Place legend on axis subplots
    style : list or dict
        matplotlib line style per column
    logx : boolean, default False
        Use log scaling on x axis
    logy : boolean, default False
        Use log scaling on y axis
    loglog : boolean, default False
        Use log scaling on both x and y axes
    xticks : sequence
        Values to use for the xticks
    yticks : sequence
        Values to use for the yticks
    xlim : 2-tuple/list
    ylim : 2-tuple/list
    rot : int, default None
        Rotation for ticks (xticks for vertical, yticks for horizontal plots)
    fontsize : int, default None
        Font size for xticks and yticks
    colormap : str or matplotlib colormap object, default None
        Colormap to select colors from. If string, load colormap with that name
        from matplotlib.
    colorbar : boolean, optional
        If True, plot colorbar (only relevant for 'scatter' and 'hexbin' plots)
    position : float
        Specify relative alignments for bar plot layout.
        From 0 (left/bottom-end) to 1 (right/top-end). Default is 0.5 (center)
    table : boolean, Series or DataFrame, default False
        If True, draw a table using the data in the DataFrame and the data will
        be transposed to meet matplotlib's default layout.
        If a Series or DataFrame is passed, use passed data to draw a table.
    yerr : DataFrame, Series, array-like, dict and str
        See :ref:`Plotting with Error Bars <visualization.errorbars>` for
        detail.
    xerr : same types as yerr.
    label : label argument to provide to plot
    secondary_y : boolean or sequence of ints, default False
        If True then y-axis will be on the right
    mark_right : boolean, default True
        When using a secondary_y axis, automatically mark the column
        labels with "(right)" in the legend
    `**kwds` : keywords
        Options to pass to matplotlib plotting method

    Returns
    -------
    axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them

    Notes
    -----

    - See matplotlib documentation online for more on this subject
    - If `kind` = 'bar' or 'barh', you can specify relative alignments
      for bar plot layout by `position` keyword.
      From 0 (left/bottom-end) to 1 (right/top-end). Default is 0.5 (center)
    """

    # function copied from pandas.plotting._core
    # so it calls modified _plot below

    import matplotlib.pyplot as plt
    if ax is None and len(plt.get_fignums()) > 0:
        ax = _gca()
        ax = MPLPlot._get_ax_layer(ax)
    return _plot(data, kind=kind, ax=ax,
                 figsize=figsize, use_index=use_index, title=title,
                 grid=grid, legend=legend,
                 style=style, logx=logx, logy=logy, loglog=loglog,
                 xticks=xticks, yticks=yticks, xlim=xlim, ylim=ylim,
                 rot=rot, fontsize=fontsize, colormap=colormap, table=table,
                 yerr=yerr, xerr=xerr,
                 label=label, secondary_y=secondary_y,
                 **kwds)


def _plot(data, x=None, y=None, subplots=False,
          ax=None, kind='line', **kwds):
    from databricks.koalas import DataFrame
    # function copied from pandas.plotting._core
    # and adapted to handle Koalas DataFrame and Series

    kind = _get_standard_kind(kind.lower().strip())
    if kind in _all_kinds:
        klass = _plot_klass[kind]
    else:
        raise ValueError("%r is not a valid plot kind" % kind)

    # scatter and hexbin are inherited from PlanePlot which require x and y
    if kind in ('scatter', 'hexbin'):
        plot_obj = klass(data, x, y, subplots=subplots, ax=ax, kind=kind, **kwds)
    else:

        # check data type and do preprocess before applying plot
        if isinstance(data, DataFrame):
            if x is not None:
                data = data.set_index(x)
            # TODO: check if value of y is plottable
            if y is not None:
                data = data[y]

        plot_obj = klass(data, subplots=subplots, ax=ax, kind=kind, **kwds)
    plot_obj.generate()
    plot_obj.draw()
    return plot_obj.result


class KoalasSeriesPlotMethods(PandasObject):
    """
    Series plotting accessor and method.

    Plotting methods can also be accessed by calling the accessor as a method
    with the ``kind`` argument:
    ``s.plot(kind='hist')`` is equivalent to ``s.plot.hist()``
    """

    def __init__(self, data):
        self.data = data

    def __call__(self, kind='line', ax=None,
                 figsize=None, use_index=True, title=None, grid=None,
                 legend=False, style=None, logx=False, logy=False,
                 loglog=False, xticks=None, yticks=None,
                 xlim=None, ylim=None,
                 rot=None, fontsize=None, colormap=None, table=False,
                 yerr=None, xerr=None,
                 label=None, secondary_y=False, **kwds):
        return plot_series(self.data, kind=kind, ax=ax, figsize=figsize,
                           use_index=use_index, title=title, grid=grid,
                           legend=legend, style=style, logx=logx, logy=logy,
                           loglog=loglog, xticks=xticks, yticks=yticks,
                           xlim=xlim, ylim=ylim, rot=rot, fontsize=fontsize,
                           colormap=colormap, table=table, yerr=yerr,
                           xerr=xerr, label=label, secondary_y=secondary_y,
                           **kwds)
    __call__.__doc__ = plot_series.__doc__

    def line(self, x=None, y=None, **kwargs):
        """
        Plot Series or DataFrame as lines.

        This function is useful to plot lines using DataFrame's values
        as coordinates.

        Parameters
        ----------
        x : int or str, optional
            Columns to use for the horizontal axis.
            Either the location or the label of the columns to be used.
            By default, it will use the DataFrame indices.
        y : int, str, or list of them, optional
            The values to be plotted.
            Either the location or the label of the columns to be used.
            By default, it will use the remaining DataFrame numeric columns.
        **kwds
            Keyword arguments to pass on to :meth:`DataFrame.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes` or :class:`numpy.ndarray`
            Return an ndarray when ``subplots=True``.

        See Also
        --------
        matplotlib.pyplot.plot : Plot y versus x as lines and/or markers.
        """
        return self(kind="line", x=x, y=y, **kwargs)

    def bar(self, **kwds):
        """
        Vertical bar plot.

        Parameters
        ----------
        `**kwds` : optional
            Additional keyword arguments are documented in
            :meth:`Koalas.Series.plot`.

        Returns
        -------
        axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them
        """
        return self(kind='bar', **kwds)

    def barh(self, **kwds):
        """
        Make a horizontal bar plot.

        A horizontal bar plot is a plot that presents quantitative data with
        rectangular bars with lengths proportional to the values that they
        represent. A bar plot shows comparisons among discrete categories. One
        axis of the plot shows the specific categories being compared, and the
        other axis represents a measured value.

        Parameters
        ----------
        x : label or position, default DataFrame.index
            Column to be used for categories.
        y : label or position, default All numeric columns in dataframe
            Columns to be plotted from the DataFrame.
        **kwds
            Keyword arguments to pass on to :meth:`databricks.koalas.DataFrame.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes` or numpy.ndarray of them

        See Also
        --------
        matplotlib.axes.Axes.bar : Plot a vertical bar plot using matplotlib.

        Examples
        --------
        >>> df = ks.DataFrame({'lab':['A', 'B', 'C'], 'val':[10, 30, 20]})
        >>> plot = df.val.plot.barh()
        """
        return self(kind='barh', **kwds)

    def box(self, **kwds):
        """
        Make a box plot of the DataFrame columns.

        Parameters
        ----------
        `**kwds` : optional
            Additional keyword arguments are documented in
            :meth:`Koalas.Series.plot`.

        `precision`: scalar, default = 0.01
            This argument is used by Koalas to compute approximate statistics
            for building a boxplot. Use *smaller* values to get more precise
            statistics.

        Returns
        -------
        axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them

        Notes
        -----
        There are behavior differences between Koalas and pandas.

        * Koalas computes approximate statistics - expect differences between
          pandas and Koalas boxplots, especially regarding 1st and 3rd quartiles.
        * The `whis` argument is only supported as a single number.
        * Koalas doesn't support the following argument(s).

          * `bootstrap` argument is not supported
          * `autorange` argument is not supported
        """
        return self(kind='box', **kwds)

    def hist(self, bins=10, **kwds):
        """
        Draw one histogram of the DataFrame’s columns.

        Parameters
        ----------
        bins : integer, default 10
            Number of histogram bins to be used
        `**kwds` : optional
            Additional keyword arguments are documented in
            :meth:`Koalas.Series.plot`.

        Returns
        -------
        axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them
        """
        return self(kind='hist', bins=bins, **kwds)

    def kde(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.Series', method_name='kde')()

    density = kde

    def area(self, **kwds):
        """
        Draw a stacked area plot.

        An area plot displays quantitative data visually.
        This function wraps the matplotlib area function.

        Parameters
        ----------
        x : label or position, optional
            Coordinates for the X axis. By default uses the index.
        y : label or position, optional
            Column to plot. By default uses all columns.
        stacked : bool, default True
            Area plots are stacked by default. Set to False to create a
            unstacked plot.
        **kwds : optional
            Additional keyword arguments are documented in
            :meth:`DataFrame.plot`.

        Returns
        -------
        matplotlib.axes.Axes or numpy.ndarray
            Area plot, or array of area plots if subplots is True.

        Examples
        --------
        >>> df = ks.DataFrame({
        ...     'sales': [3, 2, 3, 9, 10, 6],
        ...     'signups': [5, 5, 6, 12, 14, 13],
        ...     'visits': [20, 42, 28, 62, 81, 50],
        ... }, index=pd.date_range(start='2018/01/01', end='2018/07/01',
        ...                        freq='M'))
        >>> plot = df.sales.plot.area()
        """
        return self(kind='area', **kwds)

    def pie(self, **kwds):
        """
        Generate a pie plot.

        A pie plot is a proportional representation of the numerical data in a
        column. This function wraps :meth:`matplotlib.pyplot.pie` for the
        specified column. If no column reference is passed and
        ``subplots=True`` a pie plot is drawn for each numerical column
        independently.

        Parameters
        ----------
        y : int or label, optional
            Label or position of the column to plot.
            If not provided, ``subplots=True`` argument must be passed.
        **kwds
            Keyword arguments to pass on to :meth:`Koalas.Series.plot`.

        Returns
        -------
        matplotlib.axes.Axes or np.ndarray of them
            A NumPy array is returned when `subplots` is True.

        Examples
        --------
        >>> df = ks.DataFrame({'mass': [0.330, 4.87 , 5.97],
        ...                    'radius': [2439.7, 6051.8, 6378.1]},
        ...                   index=['Mercury', 'Venus', 'Earth'])
        >>> plot = df.mass.plot.pie(figsize=(5, 5))

        >>> plot = df.mass.plot.pie(subplots=True, figsize=(6, 3))
        """
        return self(kind='pie', **kwds)


class KoalasFramePlotMethods(PandasObject):
    # TODO: not sure if Koalas wanna combine plot method for Series and DataFrame
    """
    DataFrame plotting accessor and method.

    Plotting methods can also be accessed by calling the accessor as a method
    with the ``kind`` argument:
    ``df.plot(kind='hist')`` is equivalent to ``df.plot.hist()``
    """
    def __init__(self, data):
        self.data = data

    def __call__(self, x=None, y=None, kind='line', ax=None,
                 subplots=None, sharex=None, sharey=False, layout=None,
                 figsize=None, use_index=True, title=None, grid=None,
                 legend=True, style=None, logx=False, logy=False,
                 loglog=False, xticks=None, yticks=None, xlim=None,
                 ylim=None, rot=None, fontsize=None, colormap=None,
                 table=False, yerr=None, xerr=None, secondary_y=False,
                 sort_columns=False, **kwds):
        return plot_frame(self.data, x=x, y=y, kind=kind, ax=ax,
                          subplots=subplots, sharex=sharex, sharey=sharey, layout=layout,
                          figsize=figsize, use_index=use_index, title=title, grid=grid,
                          legend=legend, style=style, logx=logx, logy=logy,
                          loglog=loglog, xticks=xticks, yticks=yticks, xlim=xlim,
                          ylim=ylim, rot=rot, fontsize=fontsize, colormap=colormap,
                          table=table, yerr=yerr, xerr=xerr, secondary_y=secondary_y,
                          sort_columns=sort_columns, **kwds)

    def line(self, x=None, y=None, **kwargs):
        """
        Plot DataFrame as lines.

        Parameters
        ----------
        x: int or str, optional
            Columns to use for the horizontal axis.
        y : int, str, or list of them, optional
            The values to be plotted.
        **kwargs
            Keyword arguments to pass on to :meth:`DataFrame.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes` or :class:`numpy.ndarray`
            Return an ndarray when ``subplots=True``.

        See Also
        --------
        matplotlib.pyplot.plot : Plot y versus x as lines and/or markers.
        """
        return self(kind='line', x=x, y=y, **kwargs)

    def kde(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.DataFrame', method_name='kde')()

    def pie(self, y=None, **kwds):
        """
        Generate a pie plot.
        A pie plot is a proportional representation of the numerical data in a
        column. This function wraps :meth:`matplotlib.pyplot.pie` for the
        specified column. If no column reference is passed and
        ``subplots=True`` a pie plot is drawn for each numerical column
        independently.

        Parameters
        ----------
        y : int or label, optional
            Label or position of the column to plot.
            If not provided, ``subplots=True`` argument must be passed.
        **kwds
            Keyword arguments to pass on to :meth:`Koalas.DataFrame.plot`.

        Returns
        -------
        matplotlib.axes.Axes or np.ndarray of them
            A NumPy array is returned when `subplots` is True.
        """
        from databricks.koalas import DataFrame

        # Pandas will raise an error if y is None and subplots if not True
        if (
                isinstance(self.data, DataFrame)
                and y is None
                and not kwds.get("subplots", False)
        ):
            raise ValueError("pie requires either y column or 'subplots=True'")
        return self(kind='pie', y=y, **kwds)

    def area(self, x=None, y=None, stacked=True, **kwds):
        """
        Draw a stacked area plot.

        An area plot displays quantitative data visually.
        This function wraps the matplotlib area function.

        Parameters
        ----------
        x : label or position, optional
            Coordinates for the X axis. By default uses the index.
        y : label or position, optional
            Column to plot. By default uses all columns.
        stacked : bool, default True
            Area plots are stacked by default. Set to False to create a
            unstacked plot.
        **kwds : optional
            Additional keyword arguments are documented in
            :meth:`DataFrame.plot`.

        Returns
        -------
        matplotlib.axes.Axes or numpy.ndarray
            Area plot, or array of area plots if subplots is True.
        """
        return self(kind='area', x=x, y=y, stacked=stacked, **kwds)

    def bar(self, x=None, y=None, **kwds):
        """
        Vertical bar plot.

        Parameters
        ----------
        x : label or position, optional
            Allows plotting of one column versus another.
            If not specified, the index of the DataFrame is used.
        y : label or position, optional
            Allows plotting of one column versus another.
            If not specified, all numerical columns are used.
        `**kwds` : optional
            Additional keyword arguments are documented in
            :meth:`Koalas.DataFrame.plot`.

        Returns
        -------
        axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them
        """
        return self(kind='bar', x=x, y=y, **kwds)

    def barh(self, x=None, y=None, **kwargs):
        """
        Make a horizontal bar plot.

        A horizontal bar plot is a plot that presents quantitative data with rectangular
        bars with lengths proportional to the values that they represent. A bar plot shows
        comparisons among discrete categories. One axis of the plot shows the specific
        categories being compared, and the other axis represents a measured value.

        Parameters
        ----------
        x : label or position, default DataFrame.index
            Column to be used for categories.
        y : label or position, default All numeric columns in dataframe
            Columns to be plotted from the DataFrame.
        **kwds:
            Keyword arguments to pass on to :meth:`databricks.koalas.DataFrame.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes` or numpy.ndarray of them

        See Also
        --------
        matplotlib.axes.Axes.bar : Plot a vertical bar plot using matplotlib.
        """
        return self(kind='barh', x=x, y=y, **kwargs)

    def hexbin(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.DataFrame', method_name='hexbin')()

    def density(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.DataFrame', method_name='density')()

    def box(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.DataFrame', method_name='box')()

    def hist(self, bw_method=None, ind=None, **kwds):
        return _unsupported_function(class_name='pd.DataFrame', method_name='hist')()

    def scatter(self, x, y, s=None, c=None, **kwds):
        """
        Create a scatter plot with varying marker point size and color.

        The coordinates of each point are defined by two dataframe columns and
        filled circles are used to represent each point. This kind of plot is
        useful to see complex correlations between two variables. Points could
        be for instance natural 2D coordinates like longitude and latitude in
        a map or, in general, any pair of metrics that can be plotted against
        each other.

        Parameters
        ----------
        x : int or str
            The column name or column position to be used as horizontal
            coordinates for each point.
        y : int or str
            The column name or column position to be used as vertical
            coordinates for each point.
        s : scalar or array_like, optional
        c : str, int or array_like, optional

        **kwds: Optional
            Keyword arguments to pass on to :meth:`databricks.koalas.DataFrame.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes` or numpy.ndarray of them

        See Also
        --------
        matplotlib.pyplot.scatter : Scatter plot using multiple input data
            formats.
        """
        return self(kind="scatter", x=x, y=y, s=s, c=c, **kwds)


def plot_frame(data, x=None, y=None, kind='line', ax=None,
               subplots=None, sharex=None, sharey=False, layout=None,
               figsize=None, use_index=True, title=None, grid=None,
               legend=True, style=None, logx=False, logy=False,
               loglog=False, xticks=None, yticks=None, xlim=None,
               ylim=None, rot=None, fontsize=None, colormap=None,
               table=False, yerr=None, xerr=None, secondary_y=False,
               sort_columns=False, **kwds):
    """
    Make plots of DataFrames using matplotlib / pylab.

    Each plot kind has a corresponding method on the
    ``DataFrame.plot`` accessor:
    ``kdf.plot(kind='line')`` is equivalent to
    ``kdf.plot.line()``.

    Parameters
    ----------
    data : DataFrame

    kind : str
        - 'line' : line plot (default)
        - 'bar' : vertical bar plot
        - 'barh' : horizontal bar plot
        - 'hist' : histogram
        - 'box' : boxplot
        - 'kde' : Kernel Density Estimation plot
        - 'density' : same as 'kde'
        - 'area' : area plot
        - 'pie' : pie plot
        - 'scatter' : scatter plot
    ax : matplotlib axes object
        If not passed, uses gca()
    x : label or position, default None
    y : label, position or list of label, positions, default None
        Allows plotting of one column versus another.
    figsize : a tuple (width, height) in inches
    use_index : boolean, default True
        Use index as ticks for x axis
    title : string or list
        Title to use for the plot. If a string is passed, print the string at
        the top of the figure. If a list is passed and `subplots` is True,
        print each item in the list above the corresponding subplot.
    grid : boolean, default None (matlab style default)
        Axis grid lines
    legend : False/True/'reverse'
        Place legend on axis subplots
    style : list or dict
        matplotlib line style per column
    logx : boolean, default False
        Use log scaling on x axis
    logy : boolean, default False
        Use log scaling on y axis
    loglog : boolean, default False
        Use log scaling on both x and y axes
    xticks : sequence
        Values to use for the xticks
    yticks : sequence
        Values to use for the yticks
    xlim : 2-tuple/list
    ylim : 2-tuple/list
    sharex: bool or None, default is None
        Whether to share x axis or not.
    sharey: bool, default is False
        Whether to share y axis or not.
    rot : int, default None
        Rotation for ticks (xticks for vertical, yticks for horizontal plots)
    fontsize : int, default None
        Font size for xticks and yticks
    colormap : str or matplotlib colormap object, default None
        Colormap to select colors from. If string, load colormap with that name
        from matplotlib.
    colorbar : boolean, optional
        If True, plot colorbar (only relevant for 'scatter' and 'hexbin' plots)
    position : float
        Specify relative alignments for bar plot layout.
        From 0 (left/bottom-end) to 1 (right/top-end). Default is 0.5 (center)
    table : boolean, Series or DataFrame, default False
        If True, draw a table using the data in the DataFrame and the data will
        be transposed to meet matplotlib's default layout.
        If a Series or DataFrame is passed, use passed data to draw a table.
    yerr : DataFrame, Series, array-like, dict and str
        See :ref:`Plotting with Error Bars <visualization.errorbars>` for
        detail.
    xerr : same types as yerr.
    label : label argument to provide to plot
    secondary_y : boolean or sequence of ints, default False
        If True then y-axis will be on the right
    mark_right : boolean, default True
        When using a secondary_y axis, automatically mark the column
        labels with "(right)" in the legend
    sort_columns: bool, default is False
        When True, will sort values on plots.
    `**kwds` : keywords
        Options to pass to matplotlib plotting method

    Returns
    -------
    axes : :class:`matplotlib.axes.Axes` or numpy.ndarray of them

    Notes
    -----

    - See matplotlib documentation online for more on this subject
    - If `kind` = 'bar' or 'barh', you can specify relative alignments
      for bar plot layout by `position` keyword.
      From 0 (left/bottom-end) to 1 (right/top-end). Default is 0.5 (center)
    """

    return _plot(data, kind=kind, x=x, y=y, ax=ax,
                 figsize=figsize, use_index=use_index, title=title,
                 grid=grid, legend=legend, subplots=subplots,
                 style=style, logx=logx, logy=logy, loglog=loglog,
                 xticks=xticks, yticks=yticks, xlim=xlim, ylim=ylim,
                 rot=rot, fontsize=fontsize, colormap=colormap, table=table,
                 yerr=yerr, xerr=xerr, sharex=sharex, sharey=sharey,
                 secondary_y=secondary_y, layout=layout, sort_columns=sort_columns,
                 **kwds)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from databricks.koalas.missing import _unsupported_function, _unsupported_property, common


def unsupported_function(method_name, deprecated=False, reason=""):
    return _unsupported_function(class_name='pd.Index', method_name=method_name,
                                 deprecated=deprecated, reason=reason)


def unsupported_property(property_name, deprecated=False, reason=""):
    return _unsupported_property(class_name='pd.Index', property_name=property_name,
                                 deprecated=deprecated, reason=reason)


class _MissingPandasLikeIndex(object):

    # Properties
    T = unsupported_property('T')
    has_duplicates = unsupported_property('has_duplicates')
    nbytes = unsupported_property('nbytes')
    ndim = unsupported_property('ndim')
    nlevels = unsupported_property('nlevels')
    shape = unsupported_property('shape')

    # Deprecated properties
    strides = unsupported_property('strides', deprecated=True)
    data = unsupported_property('data', deprecated=True)
    itemsize = unsupported_property('itemsize', deprecated=True)
    base = unsupported_property('base', deprecated=True)
    flags = unsupported_property('flags', deprecated=True)

    # Functions
    append = unsupported_function('append')
    argmax = unsupported_function('argmax')
    argmin = unsupported_function('argmin')
    argsort = unsupported_function('argsort')
    asof = unsupported_function('asof')
    asof_locs = unsupported_function('asof_locs')
    copy = unsupported_function('copy')
    delete = unsupported_function('delete')
    difference = unsupported_function('difference')
    drop = unsupported_function('drop')
    drop_duplicates = unsupported_function('drop_duplicates')
    droplevel = unsupported_function('droplevel')
    dropna = unsupported_function('dropna')
    duplicated = unsupported_function('duplicated')
    equals = unsupported_function('equals')
    factorize = unsupported_function('factorize')
    fillna = unsupported_function('fillna')
    format = unsupported_function('format')
    get_indexer = unsupported_function('get_indexer')
    get_indexer_for = unsupported_function('get_indexer_for')
    get_indexer_non_unique = unsupported_function('get_indexer_non_unique')
    get_level_values = unsupported_function('get_level_values')
    get_loc = unsupported_function('get_loc')
    get_slice_bound = unsupported_function('get_slice_bound')
    get_value = unsupported_function('get_value')
    groupby = unsupported_function('groupby')
    holds_integer = unsupported_function('holds_integer')
    identical = unsupported_function('identical')
    insert = unsupported_function('insert')
    intersection = unsupported_function('intersection')
    is_ = unsupported_function('is_')
    is_boolean = unsupported_function('is_boolean')
    is_categorical = unsupported_function('is_categorical')
    is_floating = unsupported_function('is_floating')
    is_integer = unsupported_function('is_integer')
    is_interval = unsupported_function('is_interval')
    is_lexsorted_for_tuple = unsupported_function('is_lexsorted_for_tuple')
    is_mixed = unsupported_function('is_mixed')
    is_numeric = unsupported_function('is_numeric')
    is_object = unsupported_function('is_object')
    is_type_compatible = unsupported_function('is_type_compatible')
    join = unsupported_function('join')
    map = unsupported_function('map')
    max = unsupported_function('max')
    min = unsupported_function('min')
    nunique = unsupported_function('nunique')
    putmask = unsupported_function('putmask')
    ravel = unsupported_function('ravel')
    reindex = unsupported_function('reindex')
    repeat = unsupported_function('repeat')
    searchsorted = unsupported_function('searchsorted')
    set_names = unsupported_function('set_names')
    set_value = unsupported_function('set_value')
    slice_indexer = unsupported_function('slice_indexer')
    slice_locs = unsupported_function('slice_locs')
    sort = unsupported_function('sort')
    sort_values = unsupported_function('sort_values')
    sortlevel = unsupported_function('sortlevel')
    symmetric_difference = unsupported_function('symmetric_difference')
    take = unsupported_function('take')
    to_flat_index = unsupported_function('to_flat_index')
    to_frame = unsupported_function('to_frame')
    to_native_types = unsupported_function('to_native_types')
    to_numpy = unsupported_function('to_numpy')
    transpose = unsupported_function('transpose')
    union = unsupported_function('union')
    unique = unsupported_function('unique')
    value_counts = unsupported_function('value_counts')
    view = unsupported_function('view')
    where = unsupported_function('where')

    # Deprecated functions
    get_duplicates = unsupported_function('get_duplicates', deprecated=True)
    summary = unsupported_function('summary', deprecated=True)
    get_values = unsupported_function('get_values', deprecated=True)
    item = unsupported_function('item', deprecated=True)
    contains = unsupported_function('contains', deprecated=True)

    # Properties we won't support.
    values = common.values(unsupported_property)
    array = common.array(unsupported_property)

    # Functions we won't support.
    memory_usage = common.memory_usage(unsupported_function)
    to_list = common.to_list(unsupported_function)
    tolist = common.tolist(unsupported_function)


class _MissingPandasLikeMultiIndex(object):

    # Properties
    T = unsupported_property('T')
    codes = unsupported_property('codes')
    has_duplicates = unsupported_property('has_duplicates')
    is_all_dates = unsupported_property('is_all_dates')
    levels = unsupported_property('levels')
    levshape = unsupported_property('levshape')
    ndim = unsupported_property('ndim')
    nlevels = unsupported_property('nlevels')
    shape = unsupported_property('shape')

    # Deprecated properties
    strides = unsupported_property('strides', deprecated=True)
    data = unsupported_property('data', deprecated=True)
    base = unsupported_property('base', deprecated=True)
    itemsize = unsupported_property('itemsize', deprecated=True)
    labels = unsupported_property('labels', deprecated=True)
    flags = unsupported_property('flags', deprecated=True)

    # Functions
    append = unsupported_function('append')
    argmax = unsupported_function('argmax')
    argmin = unsupported_function('argmin')
    argsort = unsupported_function('argsort')
    asof = unsupported_function('asof')
    asof_locs = unsupported_function('asof_locs')
    copy = unsupported_function('copy')
    delete = unsupported_function('delete')
    difference = unsupported_function('difference')
    drop = unsupported_function('drop')
    drop_duplicates = unsupported_function('drop_duplicates')
    droplevel = unsupported_function('droplevel')
    dropna = unsupported_function('dropna')
    duplicated = unsupported_function('duplicated')
    equal_levels = unsupported_function('equal_levels')
    equals = unsupported_function('equals')
    factorize = unsupported_function('factorize')
    fillna = unsupported_function('fillna')
    format = unsupported_function('format')
    get_indexer = unsupported_function('get_indexer')
    get_indexer_for = unsupported_function('get_indexer_for')
    get_indexer_non_unique = unsupported_function('get_indexer_non_unique')
    get_level_values = unsupported_function('get_level_values')
    get_loc = unsupported_function('get_loc')
    get_loc_level = unsupported_function('get_loc_level')
    get_locs = unsupported_function('get_locs')
    get_slice_bound = unsupported_function('get_slice_bound')
    get_value = unsupported_function('get_value')
    groupby = unsupported_function('groupby')
    holds_integer = unsupported_function('holds_integer')
    identical = unsupported_function('identical')
    insert = unsupported_function('insert')
    intersection = unsupported_function('intersection')
    is_ = unsupported_function('is_')
    is_boolean = unsupported_function('is_boolean')
    is_categorical = unsupported_function('is_categorical')
    is_floating = unsupported_function('is_floating')
    is_integer = unsupported_function('is_integer')
    is_interval = unsupported_function('is_interval')
    is_lexsorted = unsupported_function('is_lexsorted')
    is_lexsorted_for_tuple = unsupported_function('is_lexsorted_for_tuple')
    is_mixed = unsupported_function('is_mixed')
    is_numeric = unsupported_function('is_numeric')
    is_object = unsupported_function('is_object')
    is_type_compatible = unsupported_function('is_type_compatible')
    join = unsupported_function('join')
    map = unsupported_function('map')
    max = unsupported_function('max')
    min = unsupported_function('min')
    nunique = unsupported_function('nunique')
    putmask = unsupported_function('putmask')
    ravel = unsupported_function('ravel')
    reindex = unsupported_function('reindex')
    remove_unused_levels = unsupported_function('remove_unused_levels')
    reorder_levels = unsupported_function('reorder_levels')
    repeat = unsupported_function('repeat')
    searchsorted = unsupported_function('searchsorted')
    set_codes = unsupported_function('set_codes')
    set_labels = unsupported_function('set_labels')
    set_levels = unsupported_function('set_levels')
    set_names = unsupported_function('set_names')
    set_value = unsupported_function('set_value')
    slice_indexer = unsupported_function('slice_indexer')
    slice_locs = unsupported_function('slice_locs')
    sort = unsupported_function('sort')
    sort_values = unsupported_function('sort_values')
    sortlevel = unsupported_function('sortlevel')
    swaplevel = unsupported_function('swaplevel')
    symmetric_difference = unsupported_function('symmetric_difference')
    take = unsupported_function('take')
    to_flat_index = unsupported_function('to_flat_index')
    to_frame = unsupported_function('to_frame')
    to_native_types = unsupported_function('to_native_types')
    to_numpy = unsupported_function('to_numpy')
    transpose = unsupported_function('transpose')
    truncate = unsupported_function('truncate')
    union = unsupported_function('union')
    unique = unsupported_function('unique')
    value_counts = unsupported_function('value_counts')
    view = unsupported_function('view')
    where = unsupported_function('where')

    # Deprecated functions
    get_duplicates = unsupported_function('get_duplicates', deprecated=True)
    summary = unsupported_function('summary', deprecated=True)
    to_hierarchical = unsupported_function('to_hierarchical', deprecated=True)
    get_values = unsupported_function('get_values', deprecated=True)
    contains = unsupported_function('contains', deprecated=True)
    item = unsupported_function('item', deprecated=True)

    # Functions we won't support.
    values = common.values(unsupported_property)
    array = common.array(unsupported_property)

    # Properties we won't support.
    memory_usage = common.memory_usage(unsupported_function)
    to_list = common.to_list(unsupported_function)
    tolist = common.tolist(unsupported_function)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from databricks.koalas.missing import _unsupported_function, _unsupported_property


def unsupported_function(method_name, deprecated=False, reason=""):
    return _unsupported_function(class_name='pd.groupby.GroupBy', method_name=method_name,
                                 deprecated=deprecated, reason=reason)


def unsupported_property(property_name, deprecated=False, reason=""):
    return _unsupported_property(class_name='pd.groupby.GroupBy', property_name=property_name,
                                 deprecated=deprecated, reason=reason)


class _MissingPandasLikeDataFrameGroupBy(object):

    # Properties
    corr = unsupported_property('corr')
    corrwith = unsupported_property('corrwith')
    cov = unsupported_property('cov')
    dtypes = unsupported_property('dtypes')
    groups = unsupported_property('groups')
    hist = unsupported_property('hist')
    indices = unsupported_property('indices')
    mad = unsupported_property('mad')
    ngroups = unsupported_property('ngroups')
    plot = unsupported_property('plot')
    quantile = unsupported_property('quantile')
    skew = unsupported_property('skew')
    tshift = unsupported_property('tshift')

    # Deprecated properties
    take = unsupported_property('take', deprecated=True)

    # Functions
    boxplot = unsupported_function('boxplot')
    cumcount = unsupported_function('cumcount')
    describe = unsupported_function('describe')
    expanding = unsupported_function('expanding')
    get_group = unsupported_function('get_group')
    head = unsupported_function('head')
    median = unsupported_function('median')
    ngroup = unsupported_function('ngroup')
    nth = unsupported_function('nth')
    ohlc = unsupported_function('ohlc')
    pct_change = unsupported_function('pct_change')
    pipe = unsupported_function('pipe')
    prod = unsupported_function('prod')
    resample = unsupported_function('resample')
    rolling = unsupported_function('rolling')
    sem = unsupported_function('sem')
    shift = unsupported_function('shift')
    tail = unsupported_function('tail')


class _MissingPandasLikeSeriesGroupBy(object):

    # Properties
    corr = unsupported_property('corr')
    cov = unsupported_property('cov')
    dtype = unsupported_property('dtype')
    groups = unsupported_property('groups')
    hist = unsupported_property('hist')
    indices = unsupported_property('indices')
    is_monotonic_decreasing = unsupported_property('is_monotonic_decreasing')
    is_monotonic_increasing = unsupported_property('is_monotonic_increasing')
    mad = unsupported_property('mad')
    ngroups = unsupported_property('ngroups')
    plot = unsupported_property('plot')
    quantile = unsupported_property('quantile')
    skew = unsupported_property('skew')
    tshift = unsupported_property('tshift')
    unique = unsupported_property('unique')

    # Deprecated properties
    take = unsupported_property('take', deprecated=True)

    # Functions
    cumcount = unsupported_function('cumcount')
    describe = unsupported_function('describe')
    expanding = unsupported_function('expanding')
    get_group = unsupported_function('get_group')
    head = unsupported_function('head')
    median = unsupported_function('median')
    ngroup = unsupported_function('ngroup')
    nth = unsupported_function('nth')
    ohlc = unsupported_function('ohlc')
    pct_change = unsupported_function('pct_change')
    pipe = unsupported_function('pipe')
    prod = unsupported_function('prod')
    resample = unsupported_function('resample')
    rolling = unsupported_function('rolling')
    sem = unsupported_function('sem')
    shift = unsupported_function('shift')
    tail = unsupported_function('tail')



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
An internal immutable DataFrame with some metadata to manage indexes.
"""

from typing import Dict, List, Optional, Tuple, Union
from itertools import accumulate

import numpy as np
import pandas as pd
from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype, is_list_like
from pyspark import sql as spark
from pyspark._globals import _NoValue, _NoValueType
from pyspark.sql import functions as F, Window
from pyspark.sql.functions import PandasUDFType, pandas_udf
from pyspark.sql.types import DataType, StructField, StructType, to_arrow_type, LongType

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.
from databricks.koalas.config import get_option
from databricks.koalas.typedef import infer_pd_series_spark_type
from databricks.koalas.utils import column_index_level, default_session, lazy_property, scol_for


IndexMap = Tuple[str, Optional[str]]


class _InternalFrame(object):
    """
    The internal immutable DataFrame which manages Spark DataFrame and column names and index
    information.

    :ivar _sdf: Spark DataFrame
    :ivar _index_map: list of pair holding the Spark field names for indexes,
                       and the index name to be seen in Koalas DataFrame.
    :ivar _scol: Spark Column
    :ivar _data_columns: list of the Spark field names to be seen as columns in Koalas DataFrame.

    .. note:: this is an internal class. It is not supposed to be exposed to users and users
        should not directly access to it.

    The internal immutable DataFrame represents the index information for a DataFrame it belongs to.
    For instance, if we have a Koalas DataFrame as below, Pandas DataFrame does not store the index
    as columns.

    >>> kdf = ks.DataFrame({
    ...     'A': [1, 2, 3, 4],
    ...     'B': [5, 6, 7, 8],
    ...     'C': [9, 10, 11, 12],
    ...     'D': [13, 14, 15, 16],
    ...     'E': [17, 18, 19, 20]}, columns = ['A', 'B', 'C', 'D', 'E'])
    >>> kdf  # doctest: +NORMALIZE_WHITESPACE
       A  B   C   D   E
    0  1  5   9  13  17
    1  2  6  10  14  18
    2  3  7  11  15  19
    3  4  8  12  16  20

    However, all columns including index column are also stored in Spark DataFrame internally
    as below.

    >>> kdf._internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+

    In order to fill this gap, the current metadata is used by mapping Spark's internal column
    to Koalas' index. See the method below:

    * `sdf` represents the internal Spark DataFrame

    * `data_columns` represents non-indexing columns

    * `index_columns` represents internal index columns

    * `columns` represents all columns

    * `index_names` represents the external index name

    * `index_map` is zipped pairs of `index_columns` and `index_names`

    * `spark_df` represents Spark DataFrame derived by the metadata

    * `pandas_df` represents pandas DataFrame derived by the metadata

    >>> internal = kdf._internal
    >>> internal.sdf.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+
    >>> internal.data_columns
    ['A', 'B', 'C', 'D', 'E']
    >>> internal.index_columns
    ['__index_level_0__']
    >>> internal.columns
    ['__index_level_0__', 'A', 'B', 'C', 'D', 'E']
    >>> internal.index_names
    [None]
    >>> internal.index_map
    [('__index_level_0__', None)]
    >>> internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+
    >>> internal.pandas_df
       A  B   C   D   E
    0  1  5   9  13  17
    1  2  6  10  14  18
    2  3  7  11  15  19
    3  4  8  12  16  20

    In case that index is set to one of the existing column as below:

    >>> kdf1 = kdf.set_index("A")
    >>> kdf1  # doctest: +NORMALIZE_WHITESPACE
       B   C   D   E
    A
    1  5   9  13  17
    2  6  10  14  18
    3  7  11  15  19
    4  8  12  16  20

    >>> kdf1._internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+---+---+---+
    |  A|  B|  C|  D|  E|
    +---+---+---+---+---+
    |  1|  5|  9| 13| 17|
    |  2|  6| 10| 14| 18|
    |  3|  7| 11| 15| 19|
    |  4|  8| 12| 16| 20|
    +---+---+---+---+---+

    >>> internal = kdf1._internal
    >>> internal.sdf.show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+---+---+---+
    |  A|  B|  C|  D|  E|
    +---+---+---+---+---+
    |  1|  5|  9| 13| 17|
    |  2|  6| 10| 14| 18|
    |  3|  7| 11| 15| 19|
    |  4|  8| 12| 16| 20|
    +---+---+---+---+---+
    >>> internal.data_columns
    ['B', 'C', 'D', 'E']
    >>> internal.index_columns
    ['A']
    >>> internal.columns
    ['A', 'B', 'C', 'D', 'E']
    >>> internal.index_names
    ['A']
    >>> internal.index_map
    [('A', 'A')]
    >>> internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+---+---+---+
    |  A|  B|  C|  D|  E|
    +---+---+---+---+---+
    |  1|  5|  9| 13| 17|
    |  2|  6| 10| 14| 18|
    |  3|  7| 11| 15| 19|
    |  4|  8| 12| 16| 20|
    +---+---+---+---+---+
    >>> internal.pandas_df  # doctest: +NORMALIZE_WHITESPACE
       B   C   D   E
    A
    1  5   9  13  17
    2  6  10  14  18
    3  7  11  15  19
    4  8  12  16  20

    In case that index becomes a multi index as below:

    >>> kdf2 = kdf.set_index("A", append=True)
    >>> kdf2  # doctest: +NORMALIZE_WHITESPACE
         B   C   D   E
      A
    0 1  5   9  13  17
    1 2  6  10  14  18
    2 3  7  11  15  19
    3 4  8  12  16  20

    >>> kdf2._internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+

    >>> internal = kdf2._internal
    >>> internal.sdf.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+
    >>> internal.data_columns
    ['B', 'C', 'D', 'E']
    >>> internal.index_columns
    ['__index_level_0__', 'A']
    >>> internal.columns
    ['__index_level_0__', 'A', 'B', 'C', 'D', 'E']
    >>> internal.index_names
    [None, 'A']
    >>> internal.index_map
    [('__index_level_0__', None), ('A', 'A')]
    >>> internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+---+---+---+---+---+
    |__index_level_0__|  A|  B|  C|  D|  E|
    +-----------------+---+---+---+---+---+
    |                0|  1|  5|  9| 13| 17|
    |                1|  2|  6| 10| 14| 18|
    |                2|  3|  7| 11| 15| 19|
    |                3|  4|  8| 12| 16| 20|
    +-----------------+---+---+---+---+---+
    >>> internal.pandas_df  # doctest: +NORMALIZE_WHITESPACE
         B   C   D   E
      A
    0 1  5   9  13  17
    1 2  6  10  14  18
    2 3  7  11  15  19
    3 4  8  12  16  20

    For multi-level columns, it also holds column_index

    >>> columns = pd.MultiIndex.from_tuples([('X', 'A'), ('X', 'B'),
    ...                                      ('Y', 'C'), ('Y', 'D')])
    >>> kdf3 = ks.DataFrame([
    ...     [1, 2, 3, 4],
    ...     [5, 6, 7, 8],
    ...     [9, 10, 11, 12],
    ...     [13, 14, 15, 16],
    ...     [17, 18, 19, 20]], columns = columns)
    >>> kdf3  # doctest: +NORMALIZE_WHITESPACE
        X       Y
        A   B   C   D
    0   1   2   3   4
    1   5   6   7   8
    2   9  10  11  12
    3  13  14  15  16
    4  17  18  19  20

    >>> internal = kdf3._internal
    >>> internal.sdf.show()  # doctest: +NORMALIZE_WHITESPACE
    +-----------------+----------+----------+----------+----------+
    |__index_level_0__|('X', 'A')|('X', 'B')|('Y', 'C')|('Y', 'D')|
    +-----------------+----------+----------+----------+----------+
    |                0|         1|         2|         3|         4|
    |                1|         5|         6|         7|         8|
    |                2|         9|        10|        11|        12|
    |                3|        13|        14|        15|        16|
    |                4|        17|        18|        19|        20|
    +-----------------+----------+----------+----------+----------+
    >>> internal.data_columns
    ["('X', 'A')", "('X', 'B')", "('Y', 'C')", "('Y', 'D')"]
    >>> internal.column_index
    [('X', 'A'), ('X', 'B'), ('Y', 'C'), ('Y', 'D')]

    For series, it also holds scol to represent the column.

    >>> kseries = kdf1.B
    >>> kseries
    A
    1    5
    2    6
    3    7
    4    8
    Name: B, dtype: int64

    >>> internal = kseries._internal
    >>> internal.sdf.show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+---+---+---+
    |  A|  B|  C|  D|  E|
    +---+---+---+---+---+
    |  1|  5|  9| 13| 17|
    |  2|  6| 10| 14| 18|
    |  3|  7| 11| 15| 19|
    |  4|  8| 12| 16| 20|
    +---+---+---+---+---+
    >>> internal.scol
    Column<b'B'>
    >>> internal.data_columns
    ['B']
    >>> internal.index_columns
    ['A']
    >>> internal.columns
    ['A', 'B']
    >>> internal.index_names
    ['A']
    >>> internal.index_map
    [('A', 'A')]
    >>> internal.spark_internal_df.show()  # doctest: +NORMALIZE_WHITESPACE
    +---+---+
    |  A|  B|
    +---+---+
    |  1|  5|
    |  2|  6|
    |  3|  7|
    |  4|  8|
    +---+---+
    >>> internal.pandas_df  # doctest: +NORMALIZE_WHITESPACE
       B
    A
    1  5
    2  6
    3  7
    4  8
    """

    def __init__(self, sdf: spark.DataFrame,
                 index_map: Optional[List[IndexMap]] = None,
                 scol: Optional[spark.Column] = None,
                 data_columns: Optional[List[str]] = None,
                 column_index: Optional[List[Tuple[str]]] = None,
                 column_index_names: Optional[List[str]] = None) -> None:
        """
        Create a new internal immutable DataFrame to manage Spark DataFrame, column fields and
        index fields and names.

        :param sdf: Spark DataFrame to be managed.
        :param index_map: list of string pair
                           Each pair holds the index field name which exists in Spark fields,
                           and the index name.
        :param scol: Spark Column to be managed.
        :param data_columns: list of string
                              Field names to appear as columns. If scol is not None, this
                              argument is ignored, otherwise if this is None, calculated from sdf.
        :param column_index: list of tuples with the same length
                              The multi-level values in the tuples.
        :param column_index_names: Names for each of the index levels.
        """
        assert isinstance(sdf, spark.DataFrame)
        if index_map is None:
            # Here is when Koalas DataFrame is created directly from Spark DataFrame.
            assert "__index_level_0__" not in sdf.schema.names, \
                "Default index column should not appear in columns of the Spark DataFrame"

            # Create default index.
            index_map = [('__index_level_0__', None)]
            sdf = _InternalFrame.attach_default_index(sdf)

        assert index_map is not None
        assert all(isinstance(index_field, str)
                   and (index_name is None or isinstance(index_name, str))
                   for index_field, index_name in index_map)
        assert scol is None or isinstance(scol, spark.Column)
        assert data_columns is None or all(isinstance(col, str) for col in data_columns)

        self._sdf = sdf  # type: spark.DataFrame
        self._index_map = index_map  # type: List[IndexMap]
        self._scol = scol  # type: Optional[spark.Column]
        if scol is not None:
            self._data_columns = sdf.select(scol).columns
            column_index = None
            column_index_names = None
        elif data_columns is None:
            index_columns = set(index_column for index_column, _ in self._index_map)
            self._data_columns = [column for column in sdf.columns if column not in index_columns]
        else:
            self._data_columns = data_columns

        if column_index is None:
            self._column_index = [(col,) for col in self._data_columns]
        else:
            assert (len(column_index) == len(self._data_columns) and
                    all(isinstance(i, tuple) for i in column_index) and
                    len(set(len(i) for i in column_index)) <= 1)
            self._column_index = column_index

        if column_index_names is not None and not is_list_like(column_index_names):
            raise ValueError('Column_index_names should be list-like or None for a MultiIndex')

        if isinstance(column_index_names, list):
            if all(name is None for name in column_index_names):
                self._column_index_names = None
            else:
                self._column_index_names = column_index_names
        else:
            self._column_index_names = column_index_names

    @staticmethod
    def attach_default_index(sdf):
        """
        This method attaches a default index to Spark DataFrame. Spark does not have the index
        notion so corresponding column should be generated.
        There are several types of default index can be configured by `compute.default_index_type`.
        """
        default_index_type = get_option("compute.default_index_type")
        if default_index_type == "sequence":
            sequential_index = F.row_number().over(
                Window.orderBy(F.monotonically_increasing_id().asc())) - 1
            scols = [scol_for(sdf, column) for column in sdf.columns]
            return sdf.select(sequential_index.alias("__index_level_0__"), *scols)
        elif default_index_type == "distributed-sequence":
            # 1. Calculates counts per each partition ID. `counts` here is, for instance,
            #     {
            #         1: 83,
            #         6: 83,
            #         3: 83,
            #         ...
            #     }
            counts = map(lambda x: (x["key"], x["count"]),
                         sdf.groupby(F.spark_partition_id().alias("key")).count().collect())

            # 2. Calculates cumulative sum in an order of partition id.
            #     Note that it does not matter if partition id guarantees its order or not.
            #     We just need a one-by-one sequential id.

            # sort by partition key.
            sorted_counts = sorted(counts, key=lambda x: x[0])
            # get cumulative sum in an order of partition key.
            cumulative_counts = accumulate(map(lambda count: count[1], sorted_counts))
            # zip it with partition key.
            sums = dict(zip(map(lambda count: count[0], sorted_counts), cumulative_counts))

            # 3. Group by partition id and assign each range.
            def default_index(pdf):
                current_partition_max = sums[pdf["__spark_partition_id"].iloc[0]]
                offset = len(pdf)
                pdf["__index_level_0__"] = list(range(
                    current_partition_max - offset, current_partition_max))
                return pdf.drop(columns=["__spark_partition_id"])

            return_schema = StructType(
                [StructField("__index_level_0__", LongType())] + list(sdf.schema))
            grouped_map_func = pandas_udf(return_schema, PandasUDFType.GROUPED_MAP)(default_index)

            sdf = sdf.withColumn("__spark_partition_id", F.spark_partition_id())
            return sdf.groupBy("__spark_partition_id").apply(grouped_map_func)
        elif default_index_type == "distributed":
            scols = [scol_for(sdf, column) for column in sdf.columns]
            return sdf.select(
                F.monotonically_increasing_id().alias("__index_level_0__"), *scols)
        else:
            raise ValueError("'compute.default_index_type' should be one of 'sequence',"
                             " 'distributed-sequence' and 'distributed'")

    @lazy_property
    def _column_index_map(self) -> Dict[Tuple[str], str]:
        return dict(zip(self.column_index, self.data_columns))

    def column_name_for(self, column_name_or_index: Union[str, Tuple[str]]) -> str:
        """ Return the actual Spark column name for the given column name or index. """
        if column_name_or_index not in self._column_index_map:
            # TODO: assert column_name_or_index not in self.data_columns
            assert isinstance(column_name_or_index, str)
            return column_name_or_index
        else:
            return self._column_index_map[column_name_or_index]

    def scol_for(self, column_name_or_index: Union[str, Tuple[str]]) -> spark.Column:
        """ Return Spark Column for the given column name or index. """
        if self._scol is not None and (column_name_or_index == self._data_columns[0]
                                       or column_name_or_index == self._column_index[0]):
            return self._scol
        else:
            return scol_for(self._sdf, self.column_name_for(column_name_or_index))

    def spark_type_for(self, column_name_or_index: Union[str, Tuple[str]]) -> DataType:
        """ Return DataType for the given column name or index. """
        return self._sdf.schema[self.column_name_for(column_name_or_index)].dataType

    @property
    def sdf(self) -> spark.DataFrame:
        """ Return the managed Spark DataFrame. """
        return self._sdf

    @property
    def data_columns(self) -> List[str]:
        """ Return the managed column field names. """
        return self._data_columns

    @lazy_property
    def data_scols(self) -> List[spark.Column]:
        """ Return Spark Columns for the managed data columns. """
        return [self.scol_for(column) for column in self.column_index]

    @lazy_property
    def index_columns(self) -> List[str]:
        """ Return the managed index field names. """
        return [index_column for index_column, _ in self._index_map]

    @lazy_property
    def index_scols(self) -> List[spark.Column]:
        """ Return Spark Columns for the managed index columns. """
        return [self.scol_for(column) for column in self.index_columns]

    @lazy_property
    def columns(self) -> List[str]:
        """ Return all the field names including index field names. """
        index_columns = set(self.index_columns)
        return self.index_columns + [column for column in self._data_columns
                                     if column not in index_columns]

    @lazy_property
    def scols(self) -> List[spark.Column]:
        """ Return Spark Columns for the managed columns including index columns. """
        return [self.scol_for(column) for column in self.columns]

    @property
    def index_map(self) -> List[IndexMap]:
        """ Return the managed index information. """
        assert len(self._index_map) > 0
        return self._index_map

    @lazy_property
    def index_names(self) -> List[Optional[str]]:
        """ Return the managed index names. """
        return [index_name for _, index_name in self.index_map]

    @property
    def scol(self) -> Optional[spark.Column]:
        """ Return the managed Spark Column. """
        return self._scol

    @property
    def column_index(self) -> List[Tuple[str]]:
        """ Return the managed column index. """
        return self._column_index

    @lazy_property
    def column_index_level(self) -> int:
        """ Return the level of the column index. """
        return column_index_level(self._column_index)

    @property
    def column_index_names(self) -> Optional[List[str]]:
        """ Return names of the index levels. """
        return self._column_index_names

    @lazy_property
    def spark_internal_df(self) -> spark.DataFrame:
        """
        Return as Spark DataFrame. This contains index columns as well
        and should be only used for internal purposes.
        """
        index_columns = set(self.index_columns)
        data_columns = []
        for column, idx in zip(self._data_columns, self.column_index):
            if column not in index_columns:
                scol = self.scol_for(idx)
                name = str(idx) if len(idx) > 1 else idx[0]
                if column != name:
                    scol = scol.alias(name)
                data_columns.append(scol)
        return self._sdf.select(self.index_scols + data_columns)

    @lazy_property
    def spark_df(self) -> spark.DataFrame:
        """ Return as Spark DataFrame. """
        index_columns = set(self.index_columns)
        data_columns = []
        for column, idx in zip(self._data_columns, self.column_index):
            if column not in index_columns:
                scol = self.scol_for(idx)
                name = str(idx) if len(idx) > 1 else idx[0]
                if column != name:
                    scol = scol.alias(name)
                data_columns.append(scol)
        return self._sdf.select(data_columns)

    @lazy_property
    def pandas_df(self):
        """ Return as pandas DataFrame. """
        sdf = self.spark_internal_df
        pdf = sdf.toPandas()
        if len(pdf) == 0 and len(sdf.schema) > 0:
            pdf = pdf.astype({field.name: to_arrow_type(field.dataType).to_pandas_dtype()
                              for field in sdf.schema})

        index_columns = self.index_columns
        if len(index_columns) > 0:
            append = False
            for index_field in index_columns:
                drop = index_field not in self.data_columns
                pdf = pdf.set_index(index_field, drop=drop, append=append)
                append = True
            pdf = pdf[[str(name) if len(name) > 1 else name[0] for name in self.column_index]]

        if self.column_index_level > 1:
            pdf.columns = pd.MultiIndex.from_tuples(self._column_index)
        else:
            pdf.columns = [idx[0] for idx in self._column_index]
        if self._column_index_names is not None:
            pdf.columns.names = self._column_index_names

        index_names = self.index_names
        if len(index_names) > 0:
            pdf.index.names = index_names
        return pdf

    def copy(self, sdf: Union[spark.DataFrame, _NoValueType] = _NoValue,
             index_map: Union[List[IndexMap], _NoValueType] = _NoValue,
             scol: Union[spark.Column, _NoValueType] = _NoValue,
             data_columns: Union[List[str], _NoValueType] = _NoValue,
             column_index: Union[List[Tuple[str]], _NoValueType] = _NoValue,
             column_index_names: Union[List[str], _NoValueType] = _NoValue,) -> '_InternalFrame':
        """ Copy the immutable DataFrame.

        :param sdf: the new Spark DataFrame. If None, then the original one is used.
        :param index_map: the new index information. If None, then the original one is used.
        :param scol: the new Spark Column. If None, then the original one is used.
        :param data_columns: the new column field names. If None, then the original ones are used.
        :param column_index: the new column index.
        :param column_index_names: the new names of the index levels.
        :return: the copied immutable DataFrame.
        """
        if sdf is _NoValue:
            sdf = self._sdf
        if index_map is _NoValue:
            index_map = self._index_map
        if scol is _NoValue:
            scol = self._scol
        # FIXME(ueshin): this if-clause should be removed.
        if column_index is _NoValue:
            if data_columns is not _NoValue:
                column_index = None
            else:
                column_index = self._column_index
        if data_columns is _NoValue:
            data_columns = self._data_columns
        # FIXME(ueshin): this if-clause should be used instead of the above.
        # if column_index is _NoValue:
        #     column_index = self._column_index
        if column_index_names is _NoValue:
            column_index_names = self._column_index_names
        return _InternalFrame(sdf, index_map=index_map, scol=scol, data_columns=data_columns,
                              column_index=column_index, column_index_names=column_index_names)

    @staticmethod
    def from_pandas(pdf: pd.DataFrame) -> '_InternalFrame':
        """ Create an immutable DataFrame from pandas DataFrame.

        :param pdf: :class:`pd.DataFrame`
        :return: the created immutable DataFrame
        """
        columns = pdf.columns
        data_columns = [str(col) for col in columns]
        if isinstance(columns, pd.MultiIndex):
            column_index = columns.tolist()
        else:
            column_index = None
        column_index_names = columns.names

        index = pdf.index

        index_map = []  # type: List[IndexMap]
        if isinstance(index, pd.MultiIndex):
            if index.names is None:
                index_map = [('__index_level_{}__'.format(i), None)
                             for i in range(len(index.levels))]
            else:
                index_map = [('__index_level_{}__'.format(i) if name is None else name, name)
                             for i, name in enumerate(index.names)]
        else:
            index_map = [(index.name
                          if index.name is not None else '__index_level_0__', index.name)]

        index_columns = [index_column for index_column, _ in index_map]

        reset_index = pdf.reset_index()
        reset_index.columns = index_columns + data_columns
        schema = StructType([StructField(name, infer_pd_series_spark_type(col),
                                         nullable=bool(col.isnull().any()))
                             for name, col in reset_index.iteritems()])
        for name, col in reset_index.iteritems():
            dt = col.dtype
            if is_datetime64_dtype(dt) or is_datetime64tz_dtype(dt):
                continue
            reset_index[name] = col.replace({np.nan: None})
        sdf = default_session().createDataFrame(reset_index, schema=schema)
        return _InternalFrame(sdf=sdf, index_map=index_map, data_columns=data_columns,
                              column_index=column_index, column_index_names=column_index_names)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Base and utility classes for Koalas objects.
"""

from functools import wraps
from typing import Union

import numpy as np
import pandas as pd
from pandas.api.types import is_list_like
from pyspark import sql as spark
from pyspark.sql import functions as F, Window
from pyspark.sql.types import DoubleType, FloatType, LongType, StringType, TimestampType, \
    to_arrow_type

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.
from databricks.koalas.internal import _InternalFrame
from databricks.koalas.typedef import pandas_wraps
from databricks.koalas.utils import align_diff_series, scol_for


def _column_op(f):
    """
    A decorator that wraps APIs taking/returning Spark Column so that Koalas Series can be
    supported too. If this decorator is used for the `f` function that takes Spark Column and
    returns Spark Column, decorated `f` takes Koalas Series as well and returns Koalas
    Series.

    :param f: a function that takes Spark Column and returns Spark Column.
    :param self: Koalas Series
    :param args: arguments that the function `f` takes.
    """
    @wraps(f)
    def wrapper(self, *args):
        # It is possible for the function `f` takes other arguments than Spark Column.
        # To cover this case, explicitly check if the argument is Koalas Series and
        # extract Spark Column. For other arguments, they are used as are.
        cols = [arg for arg in args if isinstance(arg, IndexOpsMixin)]
        if all(self._kdf is col._kdf for col in cols):
            # Same DataFrame anchors
            args = [arg._scol if isinstance(arg, IndexOpsMixin) else arg for arg in args]
            scol = f(self._scol, *args)

            return self._with_new_scol(scol)
        else:
            # Different DataFrame anchors
            def apply_func(this_column, *that_columns):
                return f(this_column, *that_columns)

            return align_diff_series(apply_func, self, *args, how="full")

    return wrapper


def _numpy_column_op(f):
    @wraps(f)
    def wrapper(self, *args):
        # PySpark does not support NumPy type out of the box. For now, we convert NumPy types
        # into some primitive types understandable in PySpark.
        new_args = []
        for arg in args:
            # TODO: This is a quick hack to support NumPy type. We should revisit this.
            if isinstance(self.spark_type, LongType) and isinstance(arg, np.timedelta64):
                new_args.append(float(arg / np.timedelta64(1, 's')))
            else:
                new_args.append(arg)
        return _column_op(f)(self, *new_args)
    return wrapper


def _wrap_accessor_spark(accessor, fn, return_type=None):
    """
    Wrap an accessor property or method, e.g., Series.dt.date with a spark function.
    """
    if return_type:
        return _column_op(
            lambda col: fn(col).cast(return_type)
        )(accessor._data)
    else:
        return _column_op(fn)(accessor._data)


def _wrap_accessor_pandas(accessor, fn, return_type):
    """
    Wrap an accessor property or method, e.g, Series.dt.date with a pandas function.
    """
    return pandas_wraps(fn, return_col=return_type)(accessor._data)


class IndexOpsMixin(object):
    """common ops mixin to support a unified interface / docs for Series / Index

    Assuming there are following attributes or properties and function.

    :ivar _scol: Spark Column instance
    :type _scol: pyspark.Column
    :ivar _kdf: Parent's Koalas DataFrame
    :type _kdf: ks.DataFrame

    :ivar spark_type: Spark data type
    :type spark_type: spark.types.DataType

    def _with_new_scol(self, scol: spark.Column) -> IndexOpsMixin
        Creates new object with the new column
    """
    def __init__(self, internal: _InternalFrame, kdf):
        self._internal = internal  # type: _InternalFrame
        self._kdf = kdf

    @property
    def _scol(self):
        return self._internal.scol

    # arithmetic operators
    __neg__ = _column_op(spark.Column.__neg__)

    def __add__(self, other):
        if isinstance(self.spark_type, StringType):
            # Concatenate string columns
            if isinstance(other, IndexOpsMixin) and isinstance(other.spark_type, StringType):
                return _column_op(F.concat)(self, other)
            # Handle df['col'] + 'literal'
            elif isinstance(other, str):
                return _column_op(F.concat)(self, F.lit(other))
            else:
                raise TypeError('string addition can only be applied to string series or literals.')
        else:
            return _column_op(spark.Column.__add__)(self, other)

    def __sub__(self, other):
        # Note that timestamp subtraction casts arguments to integer. This is to mimic Pandas's
        # behaviors. Pandas returns 'timedelta64[ns]' from 'datetime64[ns]'s subtraction.
        if isinstance(other, IndexOpsMixin) and isinstance(self.spark_type, TimestampType):
            if not isinstance(other.spark_type, TimestampType):
                raise TypeError('datetime subtraction can only be applied to datetime series.')
            return self.astype('bigint') - other.astype('bigint')
        else:
            return _column_op(spark.Column.__sub__)(self, other)

    __mul__ = _column_op(spark.Column.__mul__)
    __div__ = _numpy_column_op(spark.Column.__div__)
    __truediv__ = _numpy_column_op(spark.Column.__truediv__)
    __mod__ = _column_op(spark.Column.__mod__)

    def __radd__(self, other):
        # Handle 'literal' + df['col']
        if isinstance(self.spark_type, StringType) and isinstance(other, str):
            return self._with_new_scol(F.concat(F.lit(other), self._scol))
        else:
            return _column_op(spark.Column.__radd__)(self, other)

    __rsub__ = _column_op(spark.Column.__rsub__)
    __rmul__ = _column_op(spark.Column.__rmul__)
    __rdiv__ = _numpy_column_op(spark.Column.__rdiv__)
    __rtruediv__ = _numpy_column_op(spark.Column.__rtruediv__)

    def __floordiv__(self, other):
        return self._with_new_scol(
            F.floor(_numpy_column_op(spark.Column.__div__)(self, other)._scol))

    def __rfloordiv__(self, other):
        return self._with_new_scol(
            F.floor(_numpy_column_op(spark.Column.__rdiv__)(self, other)._scol))

    __rmod__ = _column_op(spark.Column.__rmod__)
    __pow__ = _column_op(spark.Column.__pow__)
    __rpow__ = _column_op(spark.Column.__rpow__)

    # logistic operators
    __eq__ = _column_op(spark.Column.__eq__)
    __ne__ = _column_op(spark.Column.__ne__)
    __lt__ = _column_op(spark.Column.__lt__)
    __le__ = _column_op(spark.Column.__le__)
    __ge__ = _column_op(spark.Column.__ge__)
    __gt__ = _column_op(spark.Column.__gt__)

    # `and`, `or`, `not` cannot be overloaded in Python,
    # so use bitwise operators as boolean operators
    __and__ = _column_op(spark.Column.__and__)
    __or__ = _column_op(spark.Column.__or__)
    __invert__ = _column_op(spark.Column.__invert__)
    __rand__ = _column_op(spark.Column.__rand__)
    __ror__ = _column_op(spark.Column.__ror__)

    @property
    def dtype(self):
        """Return the dtype object of the underlying data.

        Examples
        --------
        >>> s = ks.Series([1, 2, 3])
        >>> s.dtype
        dtype('int64')

        >>> s = ks.Series(list('abc'))
        >>> s.dtype
        dtype('O')

        >>> s = ks.Series(pd.date_range('20130101', periods=3))
        >>> s.dtype
        dtype('<M8[ns]')

        >>> s.rename("a").to_frame().set_index("a").index.dtype
        dtype('<M8[ns]')
        """
        if type(self.spark_type) == TimestampType:
            return np.dtype('datetime64[ns]')
        else:
            return np.dtype(to_arrow_type(self.spark_type).to_pandas_dtype())

    @property
    def empty(self):
        """
        Returns true if the current object is empty. Otherwise, returns false.

        >>> ks.range(10).id.empty
        False

        >>> ks.range(0).id.empty
        True

        >>> ks.DataFrame({}, index=list('abc')).index.empty
        False
        """
        return self._kdf._sdf.rdd.isEmpty()

    @property
    def hasnans(self):
        """
        Return True if it has any missing values. Otherwise, it returns False.

        >>> ks.DataFrame({}, index=list('abc')).index.hasnans
        False

        >>> ks.Series(['a', None]).hasnans
        True

        >>> ks.Series([1.0, 2.0, np.nan]).hasnans
        True

        >>> ks.Series([1, 2, 3]).hasnans
        False

        >>> ks.Series([1, 2, 3]).rename("a").to_frame().set_index("a").index.hasnans
        False
        """
        sdf = self._kdf._sdf.select(self._scol)
        col = self._scol

        ret = sdf.select(F.max(col.isNull() | F.isnan(col))).collect()[0][0]
        return ret

    @property
    def is_monotonic(self):
        """
        Return boolean if values in the object are monotonically increasing.

        .. note:: the current implementation of is_monotonic_increasing uses Spark's
            Window without specifying partition specification. This leads to move all data into
            single partition in single machine and could cause serious
            performance degradation. Avoid this method against very large dataset.

        Returns
        -------
        is_monotonic : boolean

        Examples
        --------
        >>> ser = ks.Series(['1/1/2018', '3/1/2018', '4/1/2018'])
        >>> ser.is_monotonic
        True

        >>> df = ks.DataFrame({'dates': [None, '1/1/2018', '2/1/2018', '3/1/2018']})
        >>> df.dates.is_monotonic
        False

        >>> df.index.is_monotonic
        True

        >>> ser = ks.Series([1])
        >>> ser.is_monotonic
        True

        >>> ser = ks.Series([])
        >>> ser.is_monotonic
        True

        >>> ser.rename("a").to_frame().set_index("a").index.is_monotonic
        True
        """
        col = self._scol
        window = Window.orderBy(self._kdf._internal.index_scols).rowsBetween(-1, -1)
        return self._with_new_scol((col >= F.lag(col, 1).over(window)) & col.isNotNull()).all()

    is_monotonic_increasing = is_monotonic

    @property
    def is_monotonic_decreasing(self):
        """
        Return boolean if values in the object are monotonically decreasing.

        .. note:: the current implementation of is_monotonic_decreasing uses Spark's
            Window without specifying partition specification. This leads to move all data into
            single partition in single machine and could cause serious
            performance degradation. Avoid this method against very large dataset.

        Returns
        -------
        is_monotonic : boolean

        Examples
        --------
        >>> ser = ks.Series(['4/1/2018', '3/1/2018', '1/1/2018'])
        >>> ser.is_monotonic_decreasing
        True

        >>> df = ks.DataFrame({'dates': [None, '3/1/2018', '2/1/2018', '1/1/2018']})
        >>> df.dates.is_monotonic_decreasing
        False

        >>> df.index.is_monotonic_decreasing
        False

        >>> ser = ks.Series([1])
        >>> ser.is_monotonic_decreasing
        True

        >>> ser = ks.Series([])
        >>> ser.is_monotonic_decreasing
        True

        >>> ser.rename("a").to_frame().set_index("a").index.is_monotonic_decreasing
        True
        """
        col = self._scol
        window = Window.orderBy(self._kdf._internal.index_scols).rowsBetween(-1, -1)
        return self._with_new_scol((col <= F.lag(col, 1).over(window)) & col.isNotNull()).all()

    def astype(self, dtype):
        """
        Cast a Koalas object to a specified dtype ``dtype``.

        Parameters
        ----------
        dtype : data type
            Use a numpy.dtype or Python type to cast entire pandas object to
            the same type.

        Returns
        -------
        casted : same type as caller

        See Also
        --------
        to_datetime : Convert argument to datetime.

        Examples
        --------
        >>> ser = ks.Series([1, 2], dtype='int32')
        >>> ser
        0    1
        1    2
        Name: 0, dtype: int32

        >>> ser.astype('int64')
        0    1
        1    2
        Name: 0, dtype: int64

        >>> ser.rename("a").to_frame().set_index("a").index.astype('int64')
        Int64Index([1, 2], dtype='int64', name='a')
        """
        from databricks.koalas.typedef import as_spark_type
        spark_type = as_spark_type(dtype)
        if not spark_type:
            raise ValueError("Type {} not understood".format(dtype))
        return self._with_new_scol(self._scol.cast(spark_type))

    def isin(self, values):
        """
        Check whether `values` are contained in Series.

        Return a boolean Series showing whether each element in the Series
        matches an element in the passed sequence of `values` exactly.

        Parameters
        ----------
        values : list or set
            The sequence of values to test.

        Returns
        -------
        isin : Series (bool dtype)

        Examples
        --------
        >>> s = ks.Series(['lama', 'cow', 'lama', 'beetle', 'lama',
        ...                'hippo'], name='animal')
        >>> s.isin(['cow', 'lama'])
        0     True
        1     True
        2     True
        3    False
        4     True
        5    False
        Name: animal, dtype: bool

        Passing a single string as ``s.isin('lama')`` will raise an error. Use
        a list of one element instead:

        >>> s.isin(['lama'])
        0     True
        1    False
        2     True
        3    False
        4     True
        5    False
        Name: animal, dtype: bool

        >>> s.rename("a").to_frame().set_index("a").index.isin(['lama'])
        Index([True, False, True, False, True, False], dtype='object', name='a')
        """
        if not is_list_like(values):
            raise TypeError("only list-like objects are allowed to be passed"
                            " to isin(), you passed a [{values_type}]"
                            .format(values_type=type(values).__name__))

        return self._with_new_scol(self._scol.isin(list(values))).rename(self.name)

    def isnull(self):
        """
        Detect existing (non-missing) values.

        Return a boolean same-sized object indicating if the values are NA.
        NA values, such as None or numpy.NaN, gets mapped to True values.
        Everything else gets mapped to False values. Characters such as empty strings '' or
        numpy.inf are not considered NA values
        (unless you set pandas.options.mode.use_inf_as_na = True).

        Returns
        -------
        Series : Mask of bool values for each element in Series
            that indicates whether an element is not an NA value.

        Examples
        --------
        >>> ser = ks.Series([5, 6, np.NaN])
        >>> ser.isna()  # doctest: +NORMALIZE_WHITESPACE
        0    False
        1    False
        2     True
        Name: 0, dtype: bool

        >>> ser.rename("a").to_frame().set_index("a").index.isna()
        Index([False, False, True], dtype='object', name='a')
        """
        if isinstance(self.spark_type, (FloatType, DoubleType)):
            return self._with_new_scol(self._scol.isNull() | F.isnan(self._scol)).rename(self.name)
        else:
            return self._with_new_scol(self._scol.isNull()).rename(self.name)

    isna = isnull

    def notnull(self):
        """
        Detect existing (non-missing) values.
        Return a boolean same-sized object indicating if the values are not NA.
        Non-missing values get mapped to True.
        Characters such as empty strings '' or numpy.inf are not considered NA values
        (unless you set pandas.options.mode.use_inf_as_na = True).
        NA values, such as None or numpy.NaN, get mapped to False values.

        Returns
        -------
        Series : Mask of bool values for each element in Series
            that indicates whether an element is not an NA value.

        Examples
        --------
        Show which entries in a Series are not NA.

        >>> ser = ks.Series([5, 6, np.NaN])
        >>> ser
        0    5.0
        1    6.0
        2    NaN
        Name: 0, dtype: float64

        >>> ser.notna()
        0     True
        1     True
        2    False
        Name: 0, dtype: bool

        >>> ser.rename("a").to_frame().set_index("a").index.notna()
        Index([True, True, False], dtype='object', name='a')
        """
        return (~self.isnull()).rename(self.name)

    notna = notnull

    # TODO: axis, skipna, and many arguments should be implemented.
    def all(self, axis: Union[int, str] = 0) -> bool:
        """
        Return whether all elements are True.

        Returns True unless there at least one element within a series that is
        False or equivalent (e.g. zero or empty)

        Parameters
        ----------
        axis : {0 or 'index'}, default 0
            Indicate which axis or axes should be reduced.

            * 0 / 'index' : reduce the index, return a Series whose index is the
              original column labels.

        Examples
        --------
        >>> ks.Series([True, True]).all()
        True

        >>> ks.Series([True, False]).all()
        False

        >>> ks.Series([0, 1]).all()
        False

        >>> ks.Series([1, 2, 3]).all()
        True

        >>> ks.Series([True, True, None]).all()
        True

        >>> ks.Series([True, False, None]).all()
        False

        >>> ks.Series([]).all()
        True

        >>> ks.Series([np.nan]).all()
        True

        >>> df = ks.Series([True, False, None]).rename("a").to_frame()
        >>> df.set_index("a").index.all()
        False
        """

        if axis not in [0, 'index']:
            raise ValueError('axis should be either 0 or "index" currently.')

        sdf = self._kdf._sdf.select(self._scol)
        col = scol_for(sdf, sdf.columns[0])

        # Note that we're ignoring `None`s here for now.
        # any and every was added as of Spark 3.0
        # ret = sdf.select(F.expr("every(CAST(`%s` AS BOOLEAN))" % sdf.columns[0])).collect()[0][0]
        # Here we use min as its alternative:
        ret = sdf.select(F.min(F.coalesce(col.cast('boolean'), F.lit(True)))).collect()[0][0]
        if ret is None:
            return True
        else:
            return ret

    # TODO: axis, skipna, and many arguments should be implemented.
    def any(self, axis: Union[int, str] = 0) -> bool:
        """
        Return whether any element is True.

        Returns False unless there at least one element within a series that is
        True or equivalent (e.g. non-zero or non-empty).

        Parameters
        ----------
        axis : {0 or 'index'}, default 0
            Indicate which axis or axes should be reduced.

            * 0 / 'index' : reduce the index, return a Series whose index is the
              original column labels.

        Examples
        --------
        >>> ks.Series([False, False]).any()
        False

        >>> ks.Series([True, False]).any()
        True

        >>> ks.Series([0, 0]).any()
        False

        >>> ks.Series([0, 1, 2]).any()
        True

        >>> ks.Series([False, False, None]).any()
        False

        >>> ks.Series([True, False, None]).any()
        True

        >>> ks.Series([]).any()
        False

        >>> ks.Series([np.nan]).any()
        False

        >>> df = ks.Series([True, False, None]).rename("a").to_frame()
        >>> df.set_index("a").index.any()
        True
        """

        if axis not in [0, 'index']:
            raise ValueError('axis should be either 0 or "index" currently.')

        sdf = self._kdf._sdf.select(self._scol)
        col = scol_for(sdf, sdf.columns[0])

        # Note that we're ignoring `None`s here for now.
        # any and every was added as of Spark 3.0
        # ret = sdf.select(F.expr("any(CAST(`%s` AS BOOLEAN))" % sdf.columns[0])).collect()[0][0]
        # Here we use max as its alternative:
        ret = sdf.select(F.max(F.coalesce(col.cast('boolean'), F.lit(False)))).collect()[0][0]
        if ret is None:
            return False
        else:
            return ret

    # TODO: add frep and axis parameter
    def shift(self, periods=1, fill_value=None):
        """
        Shift Series/Index by desired number of periods.

        .. note:: the current implementation of shift uses Spark's Window without
            specifying partition specification. This leads to move all data into
            single partition in single machine and could cause serious
            performance degradation. Avoid this method against very large dataset.

        Parameters
        ----------
        periods : int
            Number of periods to shift. Can be positive or negative.
        fill_value : object, optional
            The scalar value to use for newly introduced missing values.
            The default depends on the dtype of self. For numeric data, np.nan is used.

        Returns
        -------
        Copy of input Series/Index, shifted.

        Examples
        --------
        >>> df = ks.DataFrame({'Col1': [10, 20, 15, 30, 45],
        ...                    'Col2': [13, 23, 18, 33, 48],
        ...                    'Col3': [17, 27, 22, 37, 52]},
        ...                   columns=['Col1', 'Col2', 'Col3'])

        >>> df.Col1.shift(periods=3)
        0     NaN
        1     NaN
        2     NaN
        3    10.0
        4    20.0
        Name: Col1, dtype: float64

        >>> df.Col2.shift(periods=3, fill_value=0)
        0     0
        1     0
        2     0
        3    13
        4    23
        Name: Col2, dtype: int64

        >>> df.index.shift(periods=3, fill_value=0)
        Int64Index([0, 0, 0, 0, 1], dtype='int64')
        """
        return self._shift(periods, fill_value)

    def _shift(self, periods, fill_value, part_cols=()):
        if not isinstance(periods, int):
            raise ValueError('periods should be an int; however, got [%s]' % type(periods))

        col = self._scol
        window = Window.partitionBy(*part_cols).orderBy(self._internal.index_scols)\
            .rowsBetween(-periods, -periods)
        lag_col = F.lag(col, periods).over(window)
        col = F.when(lag_col.isNull() | F.isnan(lag_col), fill_value).otherwise(lag_col)
        return self._with_new_scol(col).rename(self.name)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utilities to deal with types. This is mostly focused on python3.
"""
import typing
import datetime
from inspect import getfullargspec
from functools import wraps

import numpy as np
import pandas as pd
from pandas.api.types import is_datetime64_dtype, is_datetime64tz_dtype
import pyarrow as pa
from pyspark.sql import Column
from pyspark.sql.functions import pandas_udf
import pyspark.sql.types as types

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.


__all__ = ['pandas_wraps', 'as_spark_type',
           'as_python_type', 'infer_pd_series_spark_type']


# A column of data, with the data type.
class _Series(object):
    def __init__(self, tpe):
        self.tpe = tpe  # type: types.DataType

    def __repr__(self):
        return "_SeriesType[{}]".format(self.tpe)


class _DataFrame(object):
    def __init__(self, tpe):
        # Seems we cannot specify field names. I currently gave some default names
        # `c0, c1, ... cn`.
        self.tpe = types.StructType(
            [types.StructField("c%s" % i, tpe[i])
             for i in range(len(tpe))])  # type: types.StructType

    def __repr__(self):
        return "_DataFrameType[{}]".format(self.tpe)


# The type is a scalar type that is furthermore understood by Spark.
class _Scalar(object):
    def __init__(self, tpe):
        self.tpe = tpe  # type: types.DataType

    def __repr__(self):
        return "_ScalarType[{}]".format(self.tpe)


# The type is left unspecified or we do not know about this type.
class _Unknown(object):
    def __init__(self, tpe):
        self.tpe = tpe

    def __repr__(self):
        return "_UnknownType[{}]".format(self.tpe)


X = typing.Union[_Series, _DataFrame, _Scalar, _Unknown]


def _to_stype(tpe) -> X:
    if hasattr(tpe, "__origin__") and tpe.__origin__ == ks.Series:
        inner = as_spark_type(tpe.__args__[0])
        return _Series(inner)
    if hasattr(tpe, "__origin__") and tpe.__origin__ == ks.DataFrame:
        tuple_type = tpe.__args__[0]
        if hasattr(tuple_type, "__tuple_params__"):
            # Python 3.5.0 to 3.5.2 has '__tuple_params__' instead.
            # See https://github.com/python/cpython/blob/v3.5.2/Lib/typing.py
            parameters = getattr(tuple_type, "__tuple_params__")
        else:
            parameters = getattr(tuple_type, "__args__")
        return _DataFrame([as_spark_type(t) for t in parameters])
    inner = as_spark_type(tpe)
    if inner is None:
        return _Unknown(tpe)
    else:
        return _Scalar(inner)


# First element of the list is the python base type
_base = {
    types.StringType(): [str, 'str', 'string'],
    types.BinaryType(): [bytes],
    types.ByteType(): [np.int8, 'int8', 'byte'],
    types.ShortType(): [np.int16, 'int16', 'short'],
    types.IntegerType(): [int, 'int', np.int, np.int32],
    types.LongType(): [np.int64, 'int64', 'long', 'bigint'],
    types.FloatType(): [float, 'float', np.float],
    types.DoubleType(): [np.float64, 'float64', 'double'],
    types.TimestampType(): [np.datetime64],
    types.DateType(): [datetime.date],
    types.BooleanType(): [bool, 'boolean', 'bool', np.bool],
    types.ArrayType(types.StringType()): []
}


def _build_type_dict():
    return dict([(other_type, spark_type) for (spark_type, l) in _base.items() for other_type in l]
                + [(spark_type, spark_type) for (spark_type, _) in _base.items()])


def _build_py_type_dict():
    return dict([(spark_type, l[0]) for (spark_type, l) in _base.items() if len(l) > 0])


_known_types = _build_type_dict()

_py_conversions = _build_py_type_dict()


def as_spark_type(tpe) -> types.DataType:
    """
    Given a python type, returns the equivalent spark type.
    Accepts:
    - the built-in types in python
    - the built-in types in numpy
    - list of pairs of (field_name, type)
    - dictionaries of field_name -> type
    - python3's typing system
    :param tpe:
    :return:
    """
    return _known_types.get(tpe, None)


def as_python_type(spark_tpe):
    return _py_conversions.get(spark_tpe, None)


def infer_pd_series_spark_type(s: pd.Series) -> types.DataType:
    """Infer Spark DataType from pandas Series dtype.

    :param s: :class:`pandas.Series` to be inferred
    :return: the inferred Spark data type
    """
    dt = s.dtype
    if dt == np.dtype('object'):
        if len(s) == 0 or s.isnull().all():
            raise ValueError("can not infer schema from empty or null dataset")
        return types.from_arrow_type(pa.Array.from_pandas(s).type)
    elif is_datetime64_dtype(dt) or is_datetime64tz_dtype(dt):
        return types.TimestampType()
    else:
        return types.from_arrow_type(pa.from_numpy_dtype(dt))


def _make_fun(f: typing.Callable, return_type: types.DataType, *args, **kwargs) -> 'ks.Series':
    """
    This function calls the function f while taking into account some of the
    limitations of the pandas UDF support:
    - support for keyword arguments
    - support for scalar values (as long as they are picklable)
    - support for type hints and input checks.
    :param f: the function to call. It is expected to have field annotations (see below).
    :param return_sig: the return type
    :param args: the arguments of the function
    :param kwargs: the kwargs to pass to the function
    :return: the value of executing the function: f(*args, **kwargs)

    The way this function executes depends on the what is provided as arguments:
     - if one of the arguments is a koalas series or dataframe:
        - the function is wrapped as a Spark UDF
        - the series arguments are checked to be coming from the same original anchor
        - the non-series arguments are serialized into the spark UDF.

    The function is expected to have the following arguments:
    """
    from databricks.koalas.series import Series
    # All the arguments.
    # None for columns or the value for non-columns
    frozen_args = []  # type: typing.List[typing.Any]
    # ks.Series for columns or None for the non-columns
    col_args = []  # type: typing.List[typing.Optional[Series]]
    for arg in args:
        if isinstance(arg, Series):
            frozen_args.append(None)
            col_args.append(arg)
        elif isinstance(arg, Column):
            raise ValueError('A pyspark column was passed as an argument.'
                             ' Pass a koalas series instead')
        else:
            frozen_args.append(arg)
            col_args.append(None)

    # Value is none for kwargs that are columns, and the value otherwise
    frozen_kwargs = []  # type: typing.List[typing.Tuple[str, typing.Any]]
    # Value is a spark col for kwarg that is column, and None otherwise
    col_kwargs = []  # type: typing.List[typing.Tuple[str, Series]]
    for (key, arg) in kwargs.items():
        if isinstance(arg, Series):
            col_kwargs.append((key, arg))
        elif isinstance(arg, Column):
            raise ValueError('A pyspark column was passed as an argument.'
                             ' Pass a koalas series instead')
        else:
            frozen_kwargs.append((key, arg))

    col_args_idxs = [idx for (idx, c) in enumerate(col_args) if c is not None]
    all_indexes = (col_args_idxs + [key for (key, _) in col_kwargs])  # type: ignore
    if not all_indexes:
        # No argument is related to spark
        # The function is just called through without other considerations.
        return f(*args, **kwargs)

    # We detected some columns. They need to be wrapped in a UDF to spark.
    kdf = _get_kdf(args, kwargs)

    def clean_fun(*args2):
        assert len(args2) == len(all_indexes), \
            "Missing some inputs:{}!={}".format(all_indexes, [str(c) for c in args2])
        full_args = list(frozen_args)
        full_kwargs = dict(frozen_kwargs)
        for (arg, idx) in zip(args2, all_indexes):
            if isinstance(idx, int):
                full_args[idx] = arg
            else:
                assert isinstance(idx, str), str(idx)
                full_kwargs[idx] = arg
        return f(*full_args, **full_kwargs)

    wrapped_udf = pandas_udf(clean_fun, returnType=return_type)
    name_tokens = []
    spark_col_args = []
    for col in col_args:
        if col is not None:
            spark_col_args.append(col._scol)
            name_tokens.append(col.name)
    kw_name_tokens = []
    for (key, col) in col_kwargs:
        spark_col_args.append(col._scol)
        kw_name_tokens.append("{}={}".format(key, col.name))
    col = wrapped_udf(*spark_col_args)
    series = Series(kdf._internal.copy(scol=col), anchor=kdf)  # type: 'ks.Series'
    all_name_tokens = name_tokens + sorted(kw_name_tokens)
    name = "{}({})".format(f.__name__, ", ".join(all_name_tokens))
    series = series.astype(return_type).alias(name)
    return series


def _get_kdf(args, kwargs):
    from databricks.koalas.series import Series
    all_cols = ([arg for arg in args if isinstance(arg, Series)]
                + [arg for arg in kwargs.values() if isinstance(arg, Series)])
    assert all_cols
    # TODO: check all the anchors
    s = all_cols[0]
    return s._kdf


def pandas_wraps(function=None, return_col=None, return_scalar=None):
    """ This annotation makes a function available for Koalas.

    Spark requires more information about the return types than pandas, and sometimes more
    information is required. This annotations allows you to seamlessly write functions that
    work for both pandas and koalas.

    Examples
    --------

    Wrapping a function with python 3's type annotations:

    >>> from databricks.koalas import pandas_wraps
    >>> pdf = pd.DataFrame({"col1": [1, 2], "col2": [10, 20]}, dtype=np.int64)
    >>> df = ks.DataFrame(pdf)

    Consider a simple function that operates on pandas series of integers

    >>> def fun(col1):
    ...     return col1.apply(lambda x: x * 2)  # Arbitrary pandas code.
    >>> fun(pdf.col1)
    0    2
    1    4
    Name: col1, dtype: int64

    Koalas needs to know the return type in order to make this function accessible to Spark.
    The following function uses python built-in typing hint system to hint that this function
    returns a Series of integers:

    >>> @pandas_wraps
    ... def fun(col1) -> ks.Series[np.int64]:
    ...     return col1.apply(lambda x: x * 2)  # Arbitrary pandas code.

    This function works as before on pandas Series:

    >>> fun(pdf.col1)
    0    2
    1    4
    Name: col1, dtype: int64

    Now it also works on Koalas series:

    >>> fun(df.col1)
    0    2
    1    4
    Name: fun(col1), dtype: int64

    Alternatively, the type hint can be provided as an argument to the `pandas_wraps` decorator:

    >>> @pandas_wraps(return_col=np.int64)
    ... def fun(col1):
    ...     return col1.apply(lambda x: x * 2)  # Arbitrary pandas code.

    >>> fun(df.col1)
    0    2
    1    4
    Name: fun(col1), dtype: int64

    Unlike PySpark user-defined functions, the decorator supports arguments all of python's
    styles of arguments (named arguments, optional arguments, list and keyworded arguments).
    It will automatically distribute argument values that are not Koalas series. Here is an
    example of function with optional series arguments and non-series arguments:

    >>> @pandas_wraps(return_col=float)
    ... def fun(col1, col2 = None, arg1="x", **kwargs):
    ...    return 2.0 * col1 if arg1 == "x" else 3.0 * col1 * col2 * kwargs['col3']

    >>> fun(df.col1)
    0    2.0
    1    4.0
    Name: fun(col1), dtype: float32

    >>> fun(df.col1, col2=df.col2, arg1="y", col3=df.col2)
    0     300.0
    1    2400.0
    Name: fun(col1, col2=col2, col3=col2), dtype: float32

    Notes
    -----
    The arguments provided to the function must be picklable, or an error will be raised by Spark.
    The example below fails.

    >>> import sys
    >>> fun(df.col1, arg1=sys.stdout)  # doctest: +SKIP
    """
    def function_wrapper(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            # Extract the signature arguments from this function.
            sig_return = _infer_return_type(f, return_col, return_scalar)
            if not isinstance(sig_return, _Series):
                raise ValueError("Expected the return type of this function to be of type column,"
                                 " but found type {}".format(sig_return))
            spark_return_type = sig_return.tpe
            return _make_fun(f, spark_return_type, *args, **kwargs)
        return wrapper
    if callable(function):
        return function_wrapper(function)
    else:
        return function_wrapper


def _infer_return_type(f, return_col=None, return_scalar=None) -> X:
    """
    >>> def func() -> int:
    ...    pass
    >>> _infer_return_type(func).tpe
    IntegerType

    >>> def func() -> ks.Series[int]:
    ...    pass
    >>> _infer_return_type(func).tpe
    IntegerType

    >>> def func() -> ks.DataFrame[np.float, str]:
    ...    pass
    >>> _infer_return_type(func).tpe
    StructType(List(StructField(c0,FloatType,true),StructField(c1,StringType,true)))

    >>> def func() -> ks.DataFrame[np.float]:
    ...    pass
    >>> _infer_return_type(func).tpe
    StructType(List(StructField(c0,FloatType,true)))
    """
    spec = getfullargspec(f)
    return_sig = spec.annotations.get("return", None)

    if not (return_col or return_sig or return_scalar):
        raise ValueError(
            "Missing type information. It should either be provided as an argument to "
            "pandas_wraps, or as a python typing hint")
    if return_col is not None:
        if isinstance(return_col, ks.Series):
            return _to_stype(return_col)
        inner = as_spark_type(return_col)
        return _Series(inner)
    if return_scalar is not None:
        if isinstance(return_scalar, ks.Series):
            raise ValueError("Column return type {}, you should use 'return_col' to specify"
                             " it.".format(return_scalar))
        inner = as_spark_type(return_scalar)
        return _Scalar(inner)
    if return_sig is not None:
        return _to_stype(return_sig)
    assert False



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from databricks.koalas.version import __version__


def assert_pyspark_version():
    import logging
    pyspark_ver = None
    try:
        import pyspark
    except ImportError:
        raise ImportError('Unable to import pyspark - consider doing a pip install with [spark] '
                          'extra to install pyspark with pip')
    else:
        pyspark_ver = getattr(pyspark, '__version__')
        if pyspark_ver is None or pyspark_ver < '2.4':
            logging.warning(
                'Found pyspark version "{}" installed. pyspark>=2.4.0 is recommended.'
                .format(pyspark_ver if pyspark_ver is not None else '<unknown version>'))


assert_pyspark_version()

from databricks.koalas.frame import DataFrame
from databricks.koalas.indexes import Index, MultiIndex
from databricks.koalas.series import Series
from databricks.koalas.typedef import pandas_wraps
from databricks.koalas.config import get_option, set_option, reset_option

__all__ = ['read_csv', 'read_parquet', 'to_datetime', 'from_pandas',
           'get_dummies', 'DataFrame', 'Series', 'Index', 'MultiIndex', 'pandas_wraps',
           'sql', 'range', 'concat', 'melt', 'get_option', 'set_option', 'reset_option',
           'read_sql_table', 'read_sql_query', 'read_sql']


def _auto_patch():
    import os
    import logging

    # Attach a usage logger.
    logger_module = os.getenv("KOALAS_USAGE_LOGGER", None)
    if logger_module is not None:
        try:
            from databricks.koalas import usage_logging
            usage_logging.attach(logger_module)
        except Exception as e:
            from pyspark.util import _exception_message
            logger = logging.getLogger('databricks.koalas.usage_logger')
            logger.warning('Tried to attach usage logger `{}`, but an exception was raised: {}'
                           .format(logger_module, _exception_message(e)))

    # Autopatching is on by default.
    x = os.getenv("SPARK_KOALAS_AUTOPATCH", "true")
    if x.lower() in ("true", "1", "enabled"):
        logger = logging.getLogger('spark')
        logger.info("Patching spark automatically. You can disable it by setting "
                    "SPARK_KOALAS_AUTOPATCH=false in your environment")

        from pyspark.sql import dataframe as df
        df.DataFrame.to_koalas = DataFrame.to_koalas


_auto_patch()

# Import after the usage logger is attached.
from databricks.koalas.config import *
from databricks.koalas.namespace import *
from databricks.koalas.sql import sql



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from databricks.koalas.missing import _unsupported_function, _unsupported_property, common


def unsupported_function(method_name, deprecated=False, reason=""):
    return _unsupported_function(class_name='pd.Series', method_name=method_name,
                                 deprecated=deprecated, reason=reason)


def unsupported_property(property_name, deprecated=False, reason=""):
    return _unsupported_property(class_name='pd.Series', property_name=property_name,
                                 deprecated=deprecated, reason=reason)


class _MissingPandasLikeSeries(object):

    # Properties
    axes = unsupported_property('axes')
    iat = unsupported_property('iat')
    nbytes = unsupported_property('nbytes')

    # Deprecated properties
    blocks = unsupported_property('blocks', deprecated=True)
    ftypes = unsupported_property('ftypes', deprecated=True)
    ftype = unsupported_property('ftype', deprecated=True)
    is_copy = unsupported_property('is_copy', deprecated=True)
    ix = unsupported_property('ix', deprecated=True)
    asobject = unsupported_property('asobject', deprecated=True)
    strides = unsupported_property('strides', deprecated=True)
    imag = unsupported_property('imag', deprecated=True)
    itemsize = unsupported_property('itemsize', deprecated=True)
    data = unsupported_property('data', deprecated=True)
    base = unsupported_property('base', deprecated=True)
    flags = unsupported_property('flags', deprecated=True)

    # Functions
    agg = unsupported_function('agg')
    aggregate = unsupported_function('aggregate')
    align = unsupported_function('align')
    argsort = unsupported_function('argsort')
    asfreq = unsupported_function('asfreq')
    asof = unsupported_function('asof')
    at_time = unsupported_function('at_time')
    autocorr = unsupported_function('autocorr')
    between = unsupported_function('between')
    between_time = unsupported_function('between_time')
    bfill = unsupported_function('bfill')
    combine = unsupported_function('combine')
    combine_first = unsupported_function('combine_first')
    copy = unsupported_function('copy')
    cov = unsupported_function('cov')
    divmod = unsupported_function('divmod')
    dot = unsupported_function('dot')
    drop = unsupported_function('drop')
    drop_duplicates = unsupported_function('drop_duplicates')
    droplevel = unsupported_function('droplevel')
    duplicated = unsupported_function('duplicated')
    ewm = unsupported_function('ewm')
    expanding = unsupported_function('expanding')
    factorize = unsupported_function('factorize')
    ffill = unsupported_function('ffill')
    filter = unsupported_function('filter')
    first = unsupported_function('first')
    first_valid_index = unsupported_function('first_valid_index')
    get = unsupported_function('get')
    infer_objects = unsupported_function('infer_objects')
    interpolate = unsupported_function('interpolate')
    items = unsupported_function('items')
    iteritems = unsupported_function('iteritems')
    keys = unsupported_function('keys')
    last = unsupported_function('last')
    last_valid_index = unsupported_function('last_valid_index')
    mad = unsupported_function('mad')
    mask = unsupported_function('mask')
    mode = unsupported_function('mode')
    pct_change = unsupported_function('pct_change')
    pop = unsupported_function('pop')
    prod = unsupported_function('prod')
    product = unsupported_function('product')
    ravel = unsupported_function('ravel')
    rdivmod = unsupported_function('rdivmod')
    reindex = unsupported_function('reindex')
    reindex_like = unsupported_function('reindex_like')
    rename_axis = unsupported_function('rename_axis')
    reorder_levels = unsupported_function('reorder_levels')
    repeat = unsupported_function('repeat')
    replace = unsupported_function('replace')
    resample = unsupported_function('resample')
    rolling = unsupported_function('rolling')
    searchsorted = unsupported_function('searchsorted')
    sem = unsupported_function('sem')
    set_axis = unsupported_function('set_axis')
    slice_shift = unsupported_function('slice_shift')
    squeeze = unsupported_function('squeeze')
    swapaxes = unsupported_function('swapaxes')
    swaplevel = unsupported_function('swaplevel')
    tail = unsupported_function('tail')
    take = unsupported_function('take')
    to_hdf = unsupported_function('to_hdf')
    to_period = unsupported_function('to_period')
    to_sql = unsupported_function('to_sql')
    to_timestamp = unsupported_function('to_timestamp')
    truncate = unsupported_function('truncate')
    tshift = unsupported_function('tshift')
    tz_convert = unsupported_function('tz_convert')
    tz_localize = unsupported_function('tz_localize')
    unstack = unsupported_function('unstack')
    update = unsupported_function('update')
    view = unsupported_function('view')
    where = unsupported_function('where')
    xs = unsupported_function('xs')

    # Deprecated functions
    as_blocks = unsupported_function('as_blocks', deprecated=True)
    as_matrix = unsupported_function('as_matrix', deprecated=True)
    clip_lower = unsupported_function('clip_lower', deprecated=True)
    clip_upper = unsupported_function('clip_upper', deprecated=True)
    compress = unsupported_function('compress', deprecated=True)
    convert_objects = unsupported_function('convert_objects', deprecated=True)
    get_ftype_counts = unsupported_function('get_ftype_counts', deprecated=True)
    get_value = unsupported_function('get_value', deprecated=True)
    nonzero = unsupported_function('nonzero', deprecated=True)
    reindex_axis = unsupported_function('reindex_axis', deprecated=True)
    select = unsupported_function('select', deprecated=True)
    set_value = unsupported_function('set_value', deprecated=True)
    valid = unsupported_function('valid', deprecated=True)
    get_values = unsupported_function('get_values', deprecated=True)
    to_dense = unsupported_function('to_dense', deprecated=True)
    to_sparse = unsupported_function('to_sparse', deprecated=True)
    to_msgpack = unsupported_function('to_msgpack', deprecated=True)
    compound = unsupported_function('compound', deprecated=True)
    put = unsupported_function('put', deprecated=True)
    item = unsupported_function('item', deprecated=True)
    ptp = unsupported_function('ptp', deprecated=True)
    argmax = unsupported_function('argmax', deprecated=True)
    argmin = unsupported_function('argmin', deprecated=True)

    # Properties we won't support.
    values = common.values(unsupported_property)
    array = common.array(unsupported_property)
    real = unsupported_property(
        'real',
        reason="If you want to collect your data as an NumPy array, use 'to_numpy()' instead.")

    # Functions we won't support.
    memory_usage = common.memory_usage(unsupported_function)
    to_pickle = common.to_pickle(unsupported_function)
    to_xarray = common.to_xarray(unsupported_function)



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Commonly used utils in Koalas.
"""

import functools
from collections import OrderedDict
from typing import Callable, Dict, List, Tuple, Union

from pyspark import sql as spark
from pyspark.sql import functions as F
from pyspark.sql.types import FloatType
import pandas as pd

from databricks import koalas as ks  # For running doctests and reference resolution in PyCharm.


def combine_frames(this, *args, how="full"):
    """
    This method combines `this` DataFrame with a different `that` DataFrame or
    Series from a different DataFrame.

    It returns a DataFrame that has prefix `this_` and `that_` to distinct
    the columns names from both DataFrames

    It internally performs a join operation which can be expensive in general.
    So, if `compute.ops_on_diff_frames` option is False,
    this method throws an exception.
    """
    from databricks.koalas import Series
    from databricks.koalas import DataFrame
    from databricks.koalas.config import get_option

    if all(isinstance(arg, Series) for arg in args):
        assert all(arg._kdf is args[0]._kdf for arg in args), \
            "Currently only one different DataFrame (from given Series) is supported"
        if this is args[0]._kdf:
            return  # We don't need to combine. All series is in this.
        that = args[0]._kdf[[ser.name for ser in args]]
    elif len(args) == 1 and isinstance(args[0], DataFrame):
        assert isinstance(args[0], DataFrame)
        if this is args[0]:
            return  # We don't need to combine. `this` and `that` are same.
        that = args[0]
    else:
        raise AssertionError("args should be single DataFrame or "
                             "single/multiple Series")

    if get_option("compute.ops_on_diff_frames"):
        this_index_map = this._internal.index_map
        that_index_map = that._internal.index_map
        assert len(this_index_map) == len(that_index_map)

        join_scols = []
        merged_index_scols = []

        # If the same named index is found, that's used.
        for this_column, this_name in this_index_map:
            for that_col, that_name in that_index_map:
                if this_name == that_name:
                    # We should merge the Spark columns into one
                    # to mimic pandas' behavior.
                    this_scol = this._internal.scol_for(this_column)
                    that_scol = that._internal.scol_for(that_col)
                    join_scol = this_scol == that_scol
                    join_scols.append(join_scol)
                    merged_index_scols.append(
                        F.when(
                            this_scol.isNotNull(), this_scol
                        ).otherwise(that_scol).alias(this_column))
                    break
            else:
                raise ValueError("Index names must be exactly matched currently.")

        assert len(join_scols) > 0, "cannot join with no overlapping index names"

        joined_df = this._sdf.alias("this").join(
            that._sdf.alias("that"), on=join_scols, how=how)

        joined_df = joined_df.select(
            merged_index_scols +
            [this[idx]._scol.alias("__this_%s" % this._internal.column_name_for(idx))
             for idx in this._internal.column_index] +
            [that[idx]._scol.alias("__that_%s" % that._internal.column_name_for(idx))
             for idx in that._internal.column_index])

        index_columns = set(this._internal.index_columns)
        new_data_columns = [c for c in joined_df.columns if c not in index_columns]
        column_index = ([tuple(['this', *idx]) for idx in this._internal.column_index]
                        + [tuple(['that', *idx]) for idx in that._internal.column_index])
        column_index_names = (([None] + this._internal.column_index_names)
                              if this._internal.column_index_names is not None else None)
        return DataFrame(
            this._internal.copy(sdf=joined_df, data_columns=new_data_columns,
                                column_index=column_index, column_index_names=column_index_names))
    else:
        raise ValueError("Cannot combine column argument because "
                         "it comes from a different dataframe")


def align_diff_frames(resolve_func, this, that, fillna=True, how="full"):
    """
    This method aligns two different DataFrames with a given `func`. Columns are resolved and
    handled within the given `func`.
    To use this, `compute.ops_on_diff_frames` should be True, for now.

    :param resolve_func: Takes aligned (joined) DataFrame, the column of the current DataFrame, and
        the column of another DataFrame. It returns an iterable that produces Series.

        >>> from databricks.koalas.config import set_option, reset_option
        >>>
        >>> set_option("compute.ops_on_diff_frames", True)
        >>>
        >>> kdf1 = ks.DataFrame({'a': [9, 8, 7, 6, 5, 4, 3, 2, 1]})
        >>> kdf2 = ks.DataFrame({'a': [9, 8, 7, 6, 5, 4, 3, 2, 1]})
        >>>
        >>> def func(kdf, this_column_index, that_column_index):
        ...    kdf  # conceptually this is A + B.
        ...
        ...    # Within this function, Series from A or B can be performed against `kdf`.
        ...    this_idx = this_column_index[0]  # this is ('a',) from kdf1.
        ...    that_idx = that_column_index[0]  # this is ('a',) from kdf2.
        ...    new_series = (kdf[this_idx] - kdf[that_idx]).rename(str(this_idx))
        ...
        ...    # This new series will be placed in new DataFrame.
        ...    yield (new_series, this_idx)
        >>>
        >>>
        >>> align_diff_frames(func, kdf1, kdf2).sort_index()
           a
        0  0
        1  0
        2  0
        3  0
        4  0
        5  0
        6  0
        7  0
        8  0
        >>> reset_option("compute.ops_on_diff_frames")

    :param this: a DataFrame to align
    :param that: another DataFrame to align
    :param fillna: If True, it fills missing values in non-common columns in both `this` and `that`.
        Otherwise, it returns as are.
    :param how: join way. In addition, it affects how `resolve_func` resolves the column conflict.
        - full: `resolve_func` should resolve only common columns from 'this' and 'that' DataFrames.
            For instance, if 'this' has columns A, B, C and that has B, C, D, `this_columns` and
            'that_columns' in this function are B, C and B, C.
        - left: `resolve_func` should resolve columns including that columns.
            For instance, if 'this' has columns A, B, C and that has B, C, D, `this_columns` is
            B, C but `that_columns` are B, C, D.
    :return: Alined DataFrame
    """
    assert how == "full" or how == "left"

    this_column_index = this._internal.column_index
    that_column_index = that._internal.column_index
    common_column_index = set(this_column_index).intersection(that_column_index)

    # 1. Full outer join given two dataframes.
    combined = combine_frames(this, that, how=how)

    # 2. Apply given function to transform the columns in a batch and keep the new columns.
    combined_column_index = combined._internal.column_index

    that_columns_to_apply = []
    this_columns_to_apply = []
    additional_that_columns = []
    columns_to_keep = []
    column_index_to_keep = []

    for combined_idx in combined_column_index:
        for common_idx in common_column_index:
            if combined_idx == tuple(['this', *common_idx]):
                this_columns_to_apply.append(combined_idx)
                break
            elif combined_idx == tuple(['that', *common_idx]):
                that_columns_to_apply.append(combined_idx)
                break
        else:
            if how == "left" and \
                    combined_idx in [tuple(['that', *idx]) for idx in that_column_index]:
                # In this case, we will drop `that_columns` in `columns_to_keep` but passes
                # it later to `func`. `func` should resolve it.
                # Note that adding this into a separate list (`additional_that_columns`)
                # is intentional so that `this_columns` and `that_columns` can be paired.
                additional_that_columns.append(combined_idx)
            elif fillna:
                columns_to_keep.append(F.lit(None).cast(FloatType()).alias(str(combined_idx)))
                column_index_to_keep.append(combined_idx)
            else:
                columns_to_keep.append(combined._internal.scol_for(combined_idx))
                column_index_to_keep.append(combined_idx)

    that_columns_to_apply += additional_that_columns

    # Should extract columns to apply and do it in a batch in case
    # it adds new columns for example.
    if len(this_columns_to_apply) > 0 or len(that_columns_to_apply) > 0:
        kser_set, column_index_applied = \
            zip(*resolve_func(combined, this_columns_to_apply, that_columns_to_apply))
        columns_applied = [c._scol for c in kser_set]
        column_index_applied = list(column_index_applied)
    else:
        columns_applied = []
        column_index_applied = []

    applied = combined[columns_applied + columns_to_keep]
    applied.columns = pd.MultiIndex.from_tuples(column_index_applied + column_index_to_keep)

    # 3. Restore the names back and deduplicate columns.
    this_idxes = OrderedDict()
    # Add columns in an order of its original frame.
    for this_idx in this_column_index:
        for new_idx in applied._internal.column_index:
            if new_idx[1:] not in this_idxes and this_idx == new_idx[1:]:
                this_idxes[new_idx[1:]] = new_idx

    # After that, we will add the rest columns.
    other_idxes = OrderedDict()
    for new_idx in applied._internal.column_index:
        if new_idx[1:] not in this_idxes:
            other_idxes[new_idx[1:]] = new_idx

    kdf = applied[list(this_idxes.values()) + list(other_idxes.values())]
    kdf.columns = kdf.columns.droplevel()
    return kdf


def align_diff_series(func, this_series, *args, how="full"):
    from databricks.koalas.base import IndexOpsMixin
    from databricks.koalas.series import Series

    cols = [arg for arg in args if isinstance(arg, IndexOpsMixin)]
    combined = combine_frames(this_series.to_frame(), *cols, how=how)

    that_columns = [combined[('that', arg.name)]._scol
                    if isinstance(arg, IndexOpsMixin) else arg for arg in args]

    scol = func(combined[('this', this_series.name)]._scol, *that_columns).alias(this_series.name)

    return Series(combined._internal.copy(scol=scol), anchor=combined)


def default_session(conf=None):
    if conf is None:
        conf = dict()
    builder = spark.SparkSession.builder.appName("Koalas")
    for key, value in conf.items():
        builder = builder.config(key, value)
    return builder.getOrCreate()


def validate_arguments_and_invoke_function(pobj: Union[pd.DataFrame, pd.Series],
                                           koalas_func: Callable, pandas_func: Callable,
                                           input_args: Dict):
    """
    Invokes a pandas function.

    This is created because different versions of pandas support different parameters, and as a
    result when we code against the latest version, our users might get a confusing
    "got an unexpected keyword argument" error if they are using an older version of pandas.

    This function validates all the arguments, removes the ones that are not supported if they
    are simply the default value (i.e. most likely the user didn't explicitly specify it). It
    throws a TypeError if the user explicitly specify an argument that is not supported by the
    pandas version available.

    For example usage, look at DataFrame.to_html().

    :param pobj: the pandas DataFrame or Series to operate on
    :param koalas_func: koalas function, used to get default parameter values
    :param pandas_func: pandas function, used to check whether pandas supports all the arguments
    :param input_args: arguments to pass to the pandas function, often created by using locals().
                       Make sure locals() call is at the top of the function so it captures only
                       input parameters, rather than local variables.
    :return: whatever pandas_func returns
    """
    import inspect

    # Makes a copy since whatever passed in is likely created by locals(), and we can't delete
    # 'self' key from that.
    args = input_args.copy()
    del args['self']

    if 'kwargs' in args:
        # explode kwargs
        kwargs = args['kwargs']
        del args['kwargs']
        args = {**args, **kwargs}

    koalas_params = inspect.signature(koalas_func).parameters
    pandas_params = inspect.signature(pandas_func).parameters

    for param in koalas_params.values():
        if param.name not in pandas_params:
            if args[param.name] == param.default:
                del args[param.name]
            else:
                raise TypeError(
                    ("The pandas version [%s] available does not support parameter '%s' " +
                     "for function '%s'.") % (pd.__version__, param.name, pandas_func.__name__))

    args['self'] = pobj
    return pandas_func(**args)


def lazy_property(fn):
    """
    Decorator that makes a property lazy-evaluated.

    Copied from https://stevenloria.com/lazy-properties/
    """
    attr_name = '_lazy_' + fn.__name__

    @property
    @functools.wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


def scol_for(sdf: spark.DataFrame, column_name: str) -> spark.Column:
    """ Return Spark Column for the given column name. """
    return sdf['`{}`'.format(column_name)]


def column_index_level(column_index: List[Tuple[str]]) -> int:
    """ Return the level of the column index. """
    if len(column_index) == 0:
        return 0
    else:
        levels = set(len(idx) for idx in column_index)
        assert len(levels) == 1, levels
        return list(levels)[0]



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Tuple, TYPE_CHECKING

import numpy as np
import pandas as pd
import pyspark

from pyspark.ml.feature import VectorAssembler
from pyspark.ml.stat import Correlation


if TYPE_CHECKING:
    import databricks.koalas as ks


CORRELATION_OUTPUT_COLUMN = '_correlation_output'


def corr(kdf: 'ks.DataFrame', method: str = 'pearson') -> pd.DataFrame:
    """
    The correlation matrix of all the numerical columns of this dataframe.

    Only accepts scalar numerical values for now.

    :param kdf: the koalas dataframe.
    :param method: {'pearson', 'spearman'}
                   * pearson : standard correlation coefficient
                   * spearman : Spearman rank correlation
    :return: :class:`pandas.DataFrame`

    >>> ks.DataFrame({'A': [0, 1], 'B': [1, 0], 'C': ['x', 'y']}).corr()
         A    B
    A  1.0 -1.0
    B -1.0  1.0
    """
    assert method in ('pearson', 'spearman')
    ndf, fields = to_numeric_df(kdf)
    corr = Correlation.corr(ndf, CORRELATION_OUTPUT_COLUMN, method)
    pcorr = corr.toPandas()
    arr = pcorr.iloc[0, 0].toArray()
    arr = pd.DataFrame(arr)
    arr.columns = fields
    arr = arr.set_index(pd.Index(fields))
    return arr


def to_numeric_df(kdf: 'ks.DataFrame') -> Tuple[pyspark.sql.DataFrame, List[str]]:
    """
    Takes a dataframe and turns it into a dataframe containing a single numerical
    vector of doubles. This dataframe has a single field called '_1'.

    TODO: index is not preserved currently
    :param kdf: the koalas dataframe.
    :return: a pair of dataframe, list of strings (the name of the columns
             that were converted to numerical types)

    >>> to_numeric_df(ks.DataFrame({'A': [0, 1], 'B': [1, 0], 'C': ['x', 'y']}))
    (DataFrame[_correlation_output: vector], ['A', 'B'])
    """
    # TODO, it should be more robust.
    accepted_types = {np.dtype(dt) for dt in [np.int8, np.int16, np.int32, np.int64,
                                              np.float32, np.float64, np.bool_]}
    numeric_fields = [fname for fname in kdf._internal.data_columns
                      if kdf[fname].dtype in accepted_types]
    numeric_df = kdf._sdf.select(*[kdf._internal.scol_for(fname) for fname in numeric_fields])
    va = VectorAssembler(inputCols=numeric_fields, outputCol=CORRELATION_OUTPUT_COLUMN)
    v = va.transform(numeric_df).select(CORRELATION_OUTPUT_COLUMN)
    return v, numeric_fields
